#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import numpy as np
import pytest

from matrixprofile.preprocess import preprocess
from matrixprofile.preprocess import is_subsequence_constant
from matrixprofile.preprocess import add_noise_to_series
from matrixprofile.preprocess import impute_missing
from matrixprofile.preprocess import validate_preprocess_kwargs


def test_valid_preprocess_kwargs():
    preprocessing_kwargs = {
        'window': 5,
        'impute_method': 'median',
        'impute_direction': 'backward',
        'add_noise': False
    }

    valid_kwargs = validate_preprocess_kwargs(preprocessing_kwargs)
    assert(valid_kwargs['window'] == 5)
    assert(valid_kwargs['impute_method'] == 'median')
    assert(valid_kwargs['impute_direction'] == 'backward')
    assert(valid_kwargs['add_noise'] == False)

    preprocessing_kwargs = {
        'window': 5,
        'add_noise': False
    }

    valid_kwargs = validate_preprocess_kwargs(preprocessing_kwargs)
    assert(valid_kwargs['window'] == 5)
    assert(valid_kwargs['impute_method'] == 'mean')
    assert(valid_kwargs['impute_direction'] == 'forward')
    assert(valid_kwargs['add_noise'] == False)

    valid_kwargs = validate_preprocess_kwargs(None)
    assert(valid_kwargs == None)


def test_invalid_preprocess_kwargs():
    with pytest.raises(ValueError) as excinfo:
        validate_preprocess_kwargs(preprocessing_kwargs = 1)
        assert "The parameter 'preprocessing_kwargs' is not dict like!" \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        preprocessing_kwargs = {
            'win': 5,
            'impute_dir': 'backward',
        }
        validate_preprocess_kwargs(preprocessing_kwargs)
        assert "invalid key(s) for preprocessing_kwargs! valid key(s) should include " \
               "{'impute_direction', 'add_noise', 'impute_method', 'window'}" \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        preprocessing_kwargs = {'window': 'str'}
        valid_kwargs = validate_preprocess_kwargs(preprocessing_kwargs)
        assert "The value for preprocessing_kwargs['window'] is not an integer!" \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        preprocessing_kwargs = {'impute_method': False}
        valid_kwargs = validate_preprocess_kwargs(preprocessing_kwargs)
        assert "invalid imputation method! valid include options: mean, median, min, max" \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        preprocessing_kwargs = {'impute_direction': 5}
        valid_kwargs = validate_preprocess_kwargs(preprocessing_kwargs)
        assert "invalid imputation direction! valid include options: " \
               "forward, fwd, f, backward, bwd, b" \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        preprocessing_kwargs = {'add_noise': 'str'}
        valid_kwargs = validate_preprocess_kwargs(preprocessing_kwargs)
        assert "The value for preprocessing_kwargs['add_noise'] is not a boolean value!" \
            in str(excinfo.value)


def test_is_subsequence_constant():
    with pytest.raises(ValueError) as excinfo:
        ts = 1
        is_subsequence_constant(ts)
        assert "subsequence is not array like!" \
            in str(excinfo.value)

    ts = np.array([1, 1, 1, 1, 1, 1])
    assert(is_subsequence_constant(ts) == True)

    ts = np.array([1, 2, 1, 1, 1, 1])
    assert(is_subsequence_constant(ts) == False)


def test_add_noise_to_series():
    with pytest.raises(ValueError) as excinfo:
        ts = 1
        temp = add_noise_to_series(ts)
        assert "series is not array like!" \
            in str(excinfo.value)

    ts = np.array([1, 1, 1, 1, 1, 1, 1, 1])

    temp = add_noise_to_series(ts)
    assert(np.all((temp - ts) >= 0) and np.all((temp - ts) < 0.0000009))


def test_impute_missing():
    with pytest.raises(ValueError) as excinfo:
        ts = 1
        ts = impute_missing(ts, window=4)
        assert "ts is not array like!" \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        ts = np.array([1, 2, 3])
        ts = impute_missing(ts, window=4, method=False)
        assert "invalid imputation method! valid include options: mean, median, min, max" \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        ts = np.array([1, 2, 3])
        ts = impute_missing(ts, window=4, direction='a')
        assert "invalid imputation direction! valid include options: " \
               "forward, fwd, f, backward, bwd, b" \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        ts = np.array([1, 2, 3])
        ts = impute_missing(ts, window='str')
        assert "window is not an integer!" \
            in str(excinfo.value)

    ts = np.array([np.nan, np.nan, np.inf, np.nan, np.inf, np.inf, 4, 5, np.nan,
                   np.inf, np.nan, np.inf, np.inf, np.inf, np.inf, np.nan, 2])

    ts = impute_missing(ts, window=4, direction='b')
    assert(np.any(np.isnan(ts)) == False)
    assert(np.any(np.isinf(ts)) == False)


def test_preprocess():
    with pytest.raises(ValueError) as excinfo:
        ts = 1
        ts = preprocess(ts, window=4)
        assert "ts is not array like!" \
            in str(excinfo.value)

    ts = np.array([np.nan, np.inf, np.inf, np.nan, np.inf, 2, 3, 2, 3, 1, 2, 3, 4, 2,
                   np.nan, np.inf, 4, 2, 3, 4, 5, 6, 7, 8, 3, 4, 2, 3, 4, 5, 6, 7, 6,
                   5, 4, 3, np.nan, np.nan, np.inf, np.nan, np.inf, np.nan])

    ts = preprocess(ts, window=4)
    assert(np.any(np.isnan(ts)) == False)
    assert(np.any(np.isinf(ts)) == False)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.stomp import stomp
from matrixprofile.algorithms.skimp import skimp
from matrixprofile.visualize import visualize
from matrixprofile.visualize import plot_snippets
from matrixprofile.algorithms.snippets import snippets

def test_catch_all_visualize_invalid_structure():
    data = {}
    with pytest.raises(Exception) as e:
        visualize(data)
        assert('MatrixProfile, Pan-MatrixProfile or Statistics data structure expected!' == str(e.value))


def test_catch_all_visualize_mp_only():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    w = 4

    profile = stomp(ts, w, n_jobs=1)

    # expect only the matrix profile plot
    figures = visualize(profile)
    assert(len(figures) == 1)


def test_catch_all_visualize_mp_cmp():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    w = 4

    profile = stomp(ts, w, n_jobs=1)
    profile['cmp'] = np.arange(len(ts) - w + 1)

    figures = visualize(profile)
    assert(len(figures) == 2)


def test_catch_all_visualize_mp_av():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    w = 4

    profile = stomp(ts, w, n_jobs=1)
    profile['av'] = np.arange(len(ts) - w + 1)

    figures = visualize(profile)
    assert(len(figures) == 2)


def test_catch_all_visualize_mp_cmp_av():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    w = 4

    profile = stomp(ts, w, n_jobs=1)
    profile['cmp'] = np.arange(len(ts) - w + 1)
    profile['av'] = np.arange(len(ts) - w + 1)

    figures = visualize(profile)
    assert(len(figures) == 3)


def test_catch_all_visualize_mp_discords():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    w = 4

    profile = stomp(ts, w, n_jobs=1)
    profile['discords'] = [0, 1]

    figures = visualize(profile)
    assert(len(figures) == 2)


def test_catch_all_visualize_mp_motifs():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    w = 4

    profile = stomp(ts, w, n_jobs=1)
    profile['motifs'] = [{'motifs': [1, 1], 'neighbors': []}]

    figures = visualize(profile)
    assert(len(figures) == 3)


def test_catch_all_visualize_mp_motifs_discords():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    w = 4

    profile = stomp(ts, w, n_jobs=1)
    profile['discords'] = [0, 1]
    profile['motifs'] = [{'motifs': [1, 1], 'neighbors': []}]

    figures = visualize(profile)
    assert(len(figures) == 4)


def test_catch_all_visualize_pmp_only():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    w = [4, 5, 6]

    profile = skimp(ts, w, n_jobs=1)

    # expect only the matrix profile plot
    figures = visualize(profile)
    assert(len(figures) == 1)


def test_catch_all_visualize_pmp_discords():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    w = [4, 5, 6]

    profile = skimp(ts, w, n_jobs=1)
    profile['discords'] = [(0, 1), (0, 2)]

    figures = visualize(profile)
    assert(len(figures) == 3)


def test_catch_all_visualize_pmp_motifs():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    w = [4, 5, 6]

    profile = skimp(ts, w, n_jobs=1)
    profile['motifs'] = [{'motifs': [(1, 1)], 'neighbors': []}]

    figures = visualize(profile)
    assert(len(figures) == 3)

def test_catch_all_visualize_pmp_motifs_discords():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    w = [4, 5, 6]

    profile = skimp(ts, w, n_jobs=1)
    profile['discords'] = [(0, 1), (0, 2)]
    profile['motifs'] = [{'motifs': [(1, 1)], 'neighbors': []}]

    figures = visualize(profile)
    assert(len(figures) == 5)


def test_catch_all_stats():
    profile = {
        'class': 'Statistics',
        'ts': np.array([]),
        'window_size': 100
    }

    figures = visualize(profile)
    assert(len(figures) == 1)


def test_catch_all_visualize_snippets():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    snippet_size = 4
    snippet_num = 1

    snippet_list = snippets(ts, snippet_size, snippet_num)

    figures = plot_snippets(snippet_list, ts)
    assert (len(figures) == snippet_num)


# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate


import numpy as np

from matrixprofile import core


def pmp_top_k_discords(profile, exclusion_zone=None, k=3):
    """
    Computes the top K discords for the given Pan-MatrixProfile. The return
    values is a list of row by col indices.

    Notes
    -----
    This algorithm is written to work with Euclidean distance. If you submit
    a PMP of Pearson metrics, then it is first converted to Euclidean.

    Parameters
    ----------
    profile : dict
        Data structure from a PMP algorithm.
    exclusion_zone : int, Default window / 2
        The zone to exclude around the found discords to reduce trivial
        findings. By default we use the row-wise window / 2.
    k : int
        Maximum number of discords to find.

    Returns
    -------
    dict : profile
        A 2D array of indices. The first column corresponds to the row index
        and the second column corresponds to the column index of the 
        submitted PMP. It is placed back on the original object passed in as
        'discords' key.

    """
    if not core.is_pmp_obj(profile):
        raise ValueError('Expecting PMP data structure!')

    # this function requires euclidean distance
    # convert if the metric is pearson
    metric = profile.get('metric', None)
    pmp = profile.get('pmp', None)
    windows = profile.get('windows', None)
    
    tmp = None
    if metric == 'pearson':
        tmp = core.pearson_to_euclidean(pmp, windows)
    else:
        tmp = np.copy(pmp).astype('d')        
    
    # replace nan and infs with -infinity
    # for whatever reason numpy argmax finds infinity as max so
    # this is a way to get around it by converting to -infinity
    tmp[core.nan_inf_indices(tmp)] = -np.inf
            
    # iterate finding the max value k times or until negative
    # infinity is obtained
    found = []
    
    for _ in range(k):
        max_idx = np.unravel_index(np.argmax(tmp), tmp.shape)
        window = windows[max_idx[0]]
        
        if tmp[max_idx] == -np.inf:
            break
        
        found.append(max_idx)
        
        # apply exclusion zone
        # the exclusion zone is based on 1/2 of the window size
        # used to compute that specific matrix profile
        n = tmp[max_idx[0]].shape[0]
        if exclusion_zone is None:
            exclusion_zone = int(np.floor(window / 2))

        ez_start = np.max([0, max_idx[1] - exclusion_zone])
        ez_stop = np.min([n, max_idx[1] + exclusion_zone])
        tmp[max_idx[0]][ez_start:ez_stop] = -np.inf
    
    profile['discords'] = np.array(found)

    return profile


def mp_top_k_discords(profile, exclusion_zone=None, k=3):
    """
    Find the top K number of discords (anomalies) given a matrix profile,
    exclusion zone and the desired number of discords. The exclusion zone
    nullifies entries on the left and right side of the first and subsequent
    discords to remove non-trivial matches. More specifically, a discord found
    at location X will more than likely have additional discords to the left or
    right of it.

    Parameters
    ----------
    profile : dict
        The MatrixProfile data structure.
    exclusion_zone : int, Default mp algorithm ez
        Desired number of values to exclude on both sides of the anomaly.
    k : int
        Desired number of discords to find.

    Returns
    -------
    dict : profile
        The original input profile with an additional "discords" key containing
        the a np.ndarray of discord indices.

    """
    if not core.is_mp_obj(profile):
        raise ValueError('Expecting MP data structure!')

    found = []
    tmp = np.copy(profile.get('mp', None)).astype('d')
    n = len(tmp)

    # TODO: this is based on STOMP standards when this motif finding algorithm
    # originally came out. Should we default this to 4.0 instead? That seems
    # to be the common value now per new research.
    window_size = profile.get('w', None)
    if exclusion_zone is None:
        exclusion_zone = profile.get('ez', None)
    
    # obtain indices in ascending order
    indices = np.argsort(tmp)
    
    # created flipped view for discords
    indices = indices[::-1]

    for idx in indices:
        if not np.isinf(tmp[idx]):
            found.append(idx)

            # apply exclusion zone
            if exclusion_zone > 0:
                exclusion_zone_start = np.max([0, idx - exclusion_zone])
                exclusion_zone_end = np.min([n, idx + exclusion_zone])
                tmp[exclusion_zone_start:exclusion_zone_end] = np.inf

        if len(found) >= k:
            break


    profile['discords'] = np.array(found, dtype='int')

    return profile


def top_k_discords(profile, exclusion_zone=None, k=3):
    """
    Find the top K number of discords (anomalies) given a mp or pmp,
    exclusion zone and the desired number of discords. The exclusion zone
    nullifies entries on the left and right side of the first and subsequent
    discords to remove non-trivial matches. More specifically, a discord found
    at location X will more than likely have additional discords to the left or
    right of it.

    Parameters
    ----------
    profile : dict
        A MatrixProfile or Pan-MatrixProfile structure.
    exclusion_zone : int, Default mp algorithm ez
        Desired number of values to exclude on both sides of the anomaly.
    k : int
        Desired number of discords to find.

    Returns
    -------
    dict : profile
        The original profile object with an additional 'discords' key. Take
        note that a MatrixProfile discord contains a single value while the
        Pan-MatrixProfile contains a row and column index.

    """
    if not core.is_mp_or_pmp_obj(profile):
        raise ValueError('Expecting MP or PMP data structure!')

    cls = profile.get('class', None)
    func = None

    if cls == 'MatrixProfile':
        func = mp_top_k_discords
    elif cls == 'PMP':
        func = pmp_top_k_discords
    else:
        raise ValueError('Unsupported data structure!')

    return func(
        profile,
        exclusion_zone=exclusion_zone,
        k=k,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.stomp import stomp


def test_stomp_window_size_less_than_4():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8])
    w = 2

    with pytest.raises(ValueError) as excinfo:
        stomp(ts, w)
        assert 'window size must be at least 4.' in str(excinfo.value)


def test_stomp_window_size_too_small():
    ts = np.array([1, 2, 3, 4, 5, 6, 7, 8])
    w = 8

    with pytest.raises(ValueError) as excinfo:
        stomp(ts, w)
        assert 'Time series is too short' in str(excinfo.value)


def test_stomp_small_series_self_join_single_threaded():
    ts = np.array([0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0])
    w = 4
    desired = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0])
    desired_pi = np.array([4, 5, 6, 7, 0, 1, 2, 3, 0])

    desired_lmp = np.array([np.inf, np.inf, np.inf, 2.82842712, 0, 0, 0, 0, 0])
    desired_lpi = np.array([0, 0, 0, 0, 0, 1, 2, 3, 0])

    desired_rmp = np.array([0, 0, 0, 0, 0, 2.82842712, np.inf, np.inf, np.inf])
    desired_rpi = np.array([4, 5, 6, 7, 8, 8, 0, 0, 0])

    profile = stomp(ts, w, n_jobs=1)
    np.testing.assert_almost_equal(profile['mp'], desired)
    np.testing.assert_almost_equal(profile['pi'], desired_pi)

    np.testing.assert_almost_equal(profile['lmp'], desired_lmp)
    np.testing.assert_almost_equal(profile['lpi'], desired_lpi)

    np.testing.assert_almost_equal(profile['rmp'], desired_rmp)
    np.testing.assert_almost_equal(profile['rpi'], desired_rpi)


def test_stomp_small_series_self_join_multi_threaded():
    ts = np.array([0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0])
    w = 4
    desired = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0])
    desired_pi = np.array([4, 5, 6, 7, 0, 1, 2, 3, 0])

    desired_lmp = np.array([np.inf, np.inf, np.inf, 2.82842712, 0, 0, 0, 0, 0])
    desired_lpi = np.array([0, 0, 0, 0, 0, 1, 2, 3, 0])

    desired_rmp = np.array([0, 0, 0, 0, 0, 2.82842712, np.inf, np.inf, np.inf])
    desired_rpi = np.array([4, 5, 6, 7, 8, 8, 0, 0, 0])

    profile = stomp(ts, w, n_jobs=-1)
    np.testing.assert_almost_equal(profile['mp'], desired)
    np.testing.assert_almost_equal(profile['pi'], desired_pi)

    np.testing.assert_almost_equal(profile['lmp'], desired_lmp)
    np.testing.assert_almost_equal(profile['lpi'], desired_lpi)

    np.testing.assert_almost_equal(profile['rmp'], desired_rmp)
    np.testing.assert_almost_equal(profile['rpi'], desired_rpi)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.mpdist import mpdist
import matrixprofile

MODULE_PATH = matrixprofile.__path__[0]


def test_small_series_single_threaded():
    ts = np.array([
        1, 2, 3, 1, 2, 3, 4, 5, 6, 0, 0, 1, 
        1, 2, 2, 4, 5, 1, 1, 9
    ]).astype('d')
    query = np.array([
        0.23595094, 0.9865171, 0.1934413, 0.60880883,
        0.55174926, 0.77139988, 0.33529215, 0.63215848
    ]).astype('d')
    w = 4

    desired = 0.437690617625298
    actual = mpdist(ts, query, w, n_jobs=1)

    np.testing.assert_almost_equal(actual, desired)


def test_small_series_multi_threaded():
    ts = np.array([
        1, 2, 3, 1, 2, 3, 4, 5, 6, 0, 0, 1, 
        1, 2, 2, 4, 5, 1, 1, 9
    ]).astype('d')
    query = np.array([
        0.23595094, 0.9865171, 0.1934413, 0.60880883,
        0.55174926, 0.77139988, 0.33529215, 0.63215848
    ]).astype('d')
    w = 4

    desired = 0.437690617625298
    actual = mpdist(ts, query, w, n_jobs=-1)

    np.testing.assert_almost_equal(actual, desired)


def test_against_matlab():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    tsb = ts[199:300]
    w = 32

    desired = 1.460009659995543e-07
    actual = mpdist(ts, tsb, w, n_jobs=1)

    np.testing.assert_almost_equal(actual, desired)


def test_against_matlab_parallel():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    tsb = ts[199:300]
    w = 32

    desired = 1.460009659995543e-07
    actual = mpdist(ts, tsb, w, n_jobs=-1)

    np.testing.assert_almost_equal(actual, desired)


def test_ts_not_one_dimensional():
    ts = np.array([[1, 1], [2, 2]])
    tsb = np.arange(10)
    w = 32

    with pytest.raises(ValueError) as excinfo:
        mpdist(ts, tsb, w)
        assert('ts must be one dimensional!' == str(excinfo.value))


def test_tsb_not_one_dimensional():
    tsb = np.array([[1, 1], [2, 2]])
    ts = np.arange(10)
    w = 32

    with pytest.raises(ValueError) as excinfo:
        mpdist(ts, tsb, w)
        assert('ts_b must be one dimensional!' == str(excinfo.value))


def test_invalid_threshold():
    ts = np.arange(100)
    tsb = np.arange(100)
    w = 32
    threshold = -1
    error = 'threshold must be a float greater than 0 and less than 1'

    with pytest.raises(ValueError) as excinfo:
        mpdist(ts, tsb, w, threshold=threshold)
        assert(error == str(excinfo.value))

    threshold = 'str'
    with pytest.raises(ValueError) as excinfo:
        mpdist(ts, tsb, w, threshold=threshold)
        assert(error == str(excinfo.value))

    threshold = 1
    with pytest.raises(ValueError) as excinfo:
        mpdist(ts, tsb, w, threshold=threshold)
        assert(error == str(excinfo.value))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

__all__ = [
    'to_json',
    'from_json',
    'to_disk',
    'from_disk',
]

import json.tool

import numpy as np

from matrixprofile import core
from matrixprofile.io.protobuf.protobuf_utils import (
    to_mpf,
    from_mpf
)


# Supported file extensions
SUPPORTED_EXTS = set([
    'json',
    'mpf',
])

# Supported file formats
SUPPORTED_FORMATS = set([
    'json',
    'mpf',
])

def JSONSerializer(obj):
    """
    Default JSON serializer to write numpy arays and other non-supported
    data types.

    Borrowed from:
    https://stackoverflow.com/a/52604722
    """
    if type(obj).__module__ == np.__name__:
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return obj.item()

    raise TypeError('Unknown type:', type(obj))


def from_json(profile):
    """
    Converts a JSON formatted string into a profile data structure.

    Parameters
    ----------
    profile : str
        The profile as a JSON formatted string.

    Returns
    -------
    profile : dict_like
        A MatrixProfile or Pan-MatrixProfile data structure.
    """
    dct = json.load(profile)

    # handle pmp and convert to appropriate types
    if core.is_pmp_obj(dct):
        dct['pmp'] = np.array(dct['pmp'], dtype='float64')
        dct['pmpi'] = np.array(dct['pmpi'], dtype=int)
        dct['data']['ts'] = np.array(dct['data']['ts'], dtype='float64')
        dct['windows'] = np.array(dct['windows'], dtype=int)

    # handle mp
    elif core.is_mp_obj(dct):
        dct['mp'] = np.array(dct['mp'], dtype='float64')
        dct['pi'] = np.array(dct['pi'], dtype=int)

        has_l = isinstance(dct['lmp'], list)
        has_l = has_l and isinstance(dct['lpi'], list)

        if has_l:
            dct['lmp'] = np.array(dct['lmp'], dtype='float64')
            dct['lpi'] = np.array(dct['lpi'], dtype=int)

        has_r = isinstance(dct['rmp'], list)
        has_r = has_r and isinstance(dct['rpi'], list)
        
        if has_r:
            dct['rmp'] = np.array(dct['rmp'], dtype='float64')
            dct['rpi'] = np.array(dct['rpi'], dtype=int)
        
        dct['data']['ts'] = np.array(dct['data']['ts'], dtype='float64')

        if isinstance(dct['data']['query'], list):
            dct['data']['query'] = np.array(dct['data']['query'], dtype='float64')
    else:
        raise ValueError('File is not of type profile!')

    return dct


def to_json(profile):
    """
    Converts a given profile object into JSON format.

    Parameters
    ----------
    profile : dict_like
        A MatrixProfile or Pan-MatrixProfile data structure.

    Returns
    -------
    str :
        The profile as a JSON formatted string.
    """
    if not core.is_mp_or_pmp_obj(profile):
        raise ValueError('profile is expected to be of type MatrixProfile or PMP')

    return json.dumps(profile, default=JSONSerializer)


def add_extension_to_path(file_path, extension):
    """
    Utility function to add the file extension when it is not provided by the
    user in the file path.

    Parameters
    ----------
    file_path : str
        The file path.

    Returns
    -------
    str :
        The file path with the extension appended.
    str :
        The file format extension.
    """
    end = '.{}'.format(extension)
    if not file_path.endswith(end):
        file_path = '{}{}'.format(file_path, end)

    return file_path


def infer_file_format(file_path):
    """
    Attempts to determine the file type based on the extension. The extension
    is assumed to be the last dot suffix.

    Parameters
    ----------
    file_path : str
        The file path to infer the file format of.
    
    Returns
    -------
    str :
        A label described the file extension.
    """
    pieces = file_path.split('.')
    extension = pieces[-1].lower()

    if extension not in SUPPORTED_EXTS:
        raise RuntimeError('Unsupported file type with extension {}'.format(extension))

    return extension


def to_disk(profile, file_path, format='json'):
    """
    Writes a profile object of type MatrixProfile or PMP to disk as a JSON
    formatted file by default.

    Note
    ----
    The JSON format is human readable where as the mpf format is binary and
    cannot be read when opened in a text editor. When the file path does not
    include the extension, it is appended for you.

    Parameters
    ----------
    profile : dict_like
        A MatrixProfile or Pan-MatrixProfile data structure.
    file_path : str
        The path to write the file to.
    format : str, default json
        The format of the file to be written. Options include json, mpf
    """
    if not core.is_mp_or_pmp_obj(profile):
        raise ValueError('profile is expected to be of type MatrixProfile or PMP')

    if format not in SUPPORTED_FORMATS:
        raise ValueError('Unsupported file format {} given.'.format(format))

    file_path = add_extension_to_path(file_path, format)

    if format == 'json':
        with open(file_path, 'w') as out:
            out.write(to_json(profile))
    elif format == 'mpf':
        with open(file_path, 'wb') as out:
            out.write(to_mpf(profile))


def from_disk(file_path, format='infer'):
    """
    Reads a profile object of type MatrixProfile or PMP from disk into the
    respective object type. By default the type is inferred by the file
    extension.

    Parameters
    ----------
    file_path : str
        The path to read the file from.
    format : str, default infer
        The file format type to read from disk. Options include:
        infer, json, mpf
    
    Returns
    -------
    profile : dict_like, None
        A MatrixProfile or Pan-MatrixProfile data structure.
    """
    if format != 'infer':
        if format not in SUPPORTED_FORMATS:
            raise ValueError('format supplied {} is not supported'.format(format))
    else:
        format = infer_file_format(file_path)
    
    profile = None
    if format == 'json':
        with open(file_path) as f:
            profile = from_json(f)
    elif format == 'mpf':
        with open(file_path, 'rb') as f:
            profile = from_mpf(f.read())
    
    return profile


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms import skimp
from matrixprofile.algorithms.skimp import binary_split
from matrixprofile.algorithms.skimp import maximum_subsequence
from matrixprofile.exceptions import NoSolutionPossible


def test_binary_split_1():
    desired = [0]
    actual = binary_split(1)

    np.testing.assert_equal(actual, desired)


def test_binary_split_many():
    desired = [0, 5, 2, 7, 1, 3, 6, 8, 4, 9]
    actual = binary_split(10)

    np.testing.assert_equal(actual, desired)


def test_maximum_subsequence_36():
    np.random.seed(9999)
    ts = np.random.uniform(size=2**10)
    w = 2**5
    subq = ts[0:w]
    ts[0:w] = subq
    ts[w+100:w+100+w] = subq

    upper = maximum_subsequence(ts, 0.98)

    assert(upper == 36)


def test_maximum_subsequence_68():
    np.random.seed(9999)
    ts = np.random.uniform(size=2**10)
    w = 2**6
    subq = ts[0:w]
    ts[0:w] = subq
    ts[w+100:w+100+w] = subq

    upper = maximum_subsequence(ts, 0.98)

    assert(upper == 68)

def test_maximum_subsequence_no_windows():
    np.random.seed(9999)
    ts = np.random.uniform(size=2**10)
    w = 2**6
    subq = ts[0:w]
    ts[0:w] = subq
    ts[w+100:w+100+w] = subq

    with pytest.raises(NoSolutionPossible) as excinfo:
        upper = maximum_subsequence(ts, 1.0)
        assert 'no windows' in str(excinfo.value)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.mpx import mpx
from matrixprofile.algorithms.cympx import mpx_ab_parallel
import matrixprofile

MODULE_PATH = matrixprofile.__path__[0]


def test_mpx_small_series_self_join_euclidean_single_threaded():
    ts = np.array([0, 1, 1, 1, 0, 0, 2, 1, 0, 0, 2, 1])
    w = 4
    desired = np.array([1.9550, 1.9550, 0.8739, 0, 0, 1.9550, 0.8739, 0, 0])
    desired_pi = np.array([4, 5, 6, 7, 8, 1, 2, 3, 4])

    profile = mpx(ts, w, cross_correlation=False, n_jobs=1)
    np.testing.assert_almost_equal(profile['mp'], desired, decimal=4)
    np.testing.assert_almost_equal(profile['pi'], desired_pi)


def test_mpx_small_series_self_join_euclidean_multi_threaded():
    ts = np.array([0, 1, 1, 1, 0, 0, 2, 1, 0, 0, 2, 1])
    w = 4
    desired = np.array([1.9550, 1.9550, 0.8739, 0, 0, 1.9550, 0.8739, 0, 0])
    desired_pi = np.array([4, 5, 6, 7, 8, 1, 2, 3, 4])

    profile = mpx(ts, w, cross_correlation=False, n_jobs=-1)
    np.testing.assert_almost_equal(profile['mp'], desired, decimal=4)
    np.testing.assert_almost_equal(profile['pi'], desired_pi)


def test_mpx_small_series_self_join_pearson_single_threaded():
    ts = np.array([0, 1, 1, 1, 0, 0, 2, 1, 0, 0, 2, 1])
    w = 4
    desired = np.array([0.522232967867094, 0.522232967867094, 0.904534033733291, 1, 1, 0.522232967867094, 0.904534033733291, 1, 1])
    desired_pi = np.array([4, 5, 6, 7, 8, 1, 2, 3, 4])

    profile = mpx(ts, w, cross_correlation=True, n_jobs=1)
    np.testing.assert_almost_equal(profile['mp'], desired, decimal=4)
    np.testing.assert_almost_equal(profile['pi'], desired_pi)


def test_mpx_small_series_self_join_pearson_multi_threaded():
    ts = np.array([0, 1, 1, 1, 0, 0, 2, 1, 0, 0, 2, 1])
    w = 4
    desired = np.array([0.522232967867094, 0.522232967867094, 0.904534033733291, 1, 1, 0.522232967867094, 0.904534033733291, 1, 1])
    desired_pi = np.array([4, 5, 6, 7, 8, 1, 2, 3, 4])

    profile = mpx(ts, w, cross_correlation=True, n_jobs=-1)
    np.testing.assert_almost_equal(profile['mp'], desired, decimal=4)
    np.testing.assert_almost_equal(profile['pi'], desired_pi)


def test_mpx_small_series_similarity_join_single_threaded():
    ts = np.array([1, 2, 3, 1, 2, 3, 4, 5, 6, 0, 0, 1, 1, 2, 2, 4, 5, 1, 1, 9]).astype('d')
    query = np.array([0, 0, 1, 1, 2, 2, 4, 5]).astype('d')
    w = 4

    desired = np.array([
        2.36387589e+00, 2.82842712e+00, 2.17957574e+00, 6.40728972e-01,
        6.40728972e-01, 6.40728972e-01, 3.26103392e+00, 3.61947699e+00,
        3.39984131e+00, 0.00000000e+00, 4.21468485e-08, 0.00000000e+00,
        4.21468485e-08, 0.00000000e+00, 2.82842712e+00, 3.57109342e+00,
        1.73771570e+00
    ])
    desired_pi = np.array([0, 1, 4, 1, 1, 1, 2, 1, 4, 2, 1, 2, 3, 4, 2, 1, 3])

    profile = mpx(ts, w, cross_correlation=False, query=query, n_jobs=1)

    np.testing.assert_almost_equal(profile['mp'], desired, decimal=4)
    np.testing.assert_almost_equal(profile['pi'], desired_pi)


def test_mpx_small_series_similarity_join_multi_threaded():
    ts = np.array([1, 2, 3, 1, 2, 3, 4, 5, 6, 0, 0, 1, 1, 2, 2, 4, 5, 1, 1, 9]).astype('d')
    query = np.array([0, 0, 1, 1, 2, 2, 4, 5]).astype('d')
    w = 4

    desired = np.array([
        2.36387589e+00, 2.82842712e+00, 2.17957574e+00, 6.40728972e-01,
        6.40728972e-01, 6.40728972e-01, 3.26103392e+00, 3.61947699e+00,
        3.39984131e+00, 0.00000000e+00, 4.21468485e-08, 0.00000000e+00,
        4.21468485e-08, 0.00000000e+00, 2.82842712e+00, 3.57109342e+00,
        1.73771570e+00
    ])
    desired_pi = np.array([0, 1, 4, 1, 1, 1, 2, 1, 4, 2, 1, 2, 3, 4, 2, 1, 3])

    profile = mpx(ts, w, cross_correlation=False, query=query, n_jobs=-1)

    np.testing.assert_almost_equal(profile['mp'], desired, decimal=4)
    np.testing.assert_almost_equal(profile['pi'], desired_pi)


def test_mpx_similarity_join_matlab():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    tsb = ts[199:300]
    w = 32

    ml_mpa = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'mpx_ab_mpa.txt'))
    ml_mpb = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'mpx_ab_mpb.txt'))

    mpa, mpia, mpb, mpib = mpx_ab_parallel(ts, tsb, w, 0, 1)

    np.testing.assert_almost_equal(ml_mpa, mpa, decimal=4)
    np.testing.assert_almost_equal(ml_mpb, mpb, decimal=4)


def test_mpx_similarity_join_parallel_matlab():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    tsb = ts[199:300]
    w = 32

    ml_mpa = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'mpx_ab_mpa.txt'))
    ml_mpb = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'mpx_ab_mpb.txt'))

    mpa, mpia, mpb, mpib = mpx_ab_parallel(ts, tsb, w, 0, 2)

    np.testing.assert_almost_equal(ml_mpa, mpa, decimal=4)
    np.testing.assert_almost_equal(ml_mpb, mpb, decimal=4)


# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import math

from matrixprofile import core

from matrixprofile.preprocess import preprocess
from matrixprofile.preprocess import validate_preprocess_kwargs
from matrixprofile.discover import discords
from matrixprofile.discover import motifs
from matrixprofile.algorithms import skimp
from matrixprofile.algorithms import maximum_subsequence
from matrixprofile.algorithms import mpx
from matrixprofile.algorithms import scrimp_plus_plus
from matrixprofile import visualize


def analyze_pmp(ts, query, sample_pct, threshold, windows=None, n_jobs=1):
    """
    Computes the Pan-MatrixProfile, top 3 motifs and top 3 discords for the
    provided time series and query. Additionally, plots for the PMP, motifs
    and discords is provided.

    Parameters
    ----------
    ts : array_like
        The time series to analyze.
    query : array_like
        The query to analyze.
    sample_pct : float
        A float between 0 and 1 representing how many samples to compute for
        the PMP.
    threshold : float
        A correlation threshold between 0 and 1 that is used to compute the
        upper window. Note that this is used only when the windows is None.
    windows : array_like, default None
        Integers representing the desired windows to use during the
        computation of the PMP.
    n_jobs : int, default -1 (all cpu cores)
        The number of cpu cores to use when computing the PMP.
    
    Returns
    -------
    tuple : (profile, figures)
        A tuple with the first item being the profile and the second being an
        array of matplotlib figures.

    """
    ts = core.to_np_array(ts)

    if isinstance(threshold, type(None)):
        threshold = 0.98

    # when a threshold is passed, we compute the upper window
    profile = None
    if isinstance(windows, type(None)):
        profile = maximum_subsequence(ts, threshold, include_pmp=True, n_jobs=n_jobs)

        # determine windows to be computed
        # from 8 in steps of 2 until upper w
        start = 8
        windows = range(start, profile['upper_window'] + 1)

    # compute the pmp
    profile = skimp(ts, windows=windows, sample_pct=sample_pct,
                          pmp_obj=profile, n_jobs=n_jobs)

    # extract top motifs
    profile = motifs(profile)

    # extract top discords
    profile = discords(profile)

    # plot pmp
    figures = visualize(profile)

    return (profile, figures)


def analyze_mp_exact(ts, query, window, n_jobs=1):
    """
    Computes the exact MatrixProfile, top 3 motifs and top 3 discords for the
    provided time series and query. Additionally, the MatrixProfile, discords
    and motifs are visualized.

    Parameters
    ----------
    ts : array_like
        The time series to analyze.
    query : array_like
        The query to analyze.
    window : int
        The window size to compute the MatrixProfile.
    n_jobs : int, default -1 (all cpu cores)
        The number of cpu cores to use when computing the MP.
    
    Returns
    -------
    tuple : (profile, figures)
        A tuple with the first item being the profile and the second being an
        array of matplotlib figures.

    """
    ts = core.to_np_array(ts)

    # compute mp
    profile = mpx(ts, window, query=query, n_jobs=n_jobs)

    # extract top motifs
    profile = motifs(profile)

    # extract top discords
    profile = discords(profile)

    # plot mp
    figures = visualize(profile)

    return (profile, figures)


def analyze_mp_approximate(ts, query, window, sample_pct, n_jobs=1):
    """
    Computes the exact MatrixProfile, top 3 motifs and top 3 discords for the
    provided time series and query. Additionally, the MatrixProfile, discords
    and motifs are visualized.

    Parameters
    ----------
    ts : array_like
        The time series to analyze.
    query : array_like
        The query to analyze.
    window : int
        The window size to compute the MatrixProfile.
    sample_pct : float
        A float between 0 and 1 representing how many samples to compute for
        the MP. When it is 1, it is the same as using the exact algorithm.
    n_jobs : int, default -1 (all cpu cores)
        The number of cpu cores to use when computing the MP.
    
    Returns
    -------
    tuple : (profile, figures)
        A tuple with the first item being the profile and the second being an
        array of matplotlib figures.

    """
    ts = core.to_np_array(ts)

    # compute mp
    profile = scrimp_plus_plus(ts, window, query=query, sample_pct=sample_pct,
        n_jobs=n_jobs)

    # extract top motifs
    profile = motifs(profile)

    # extract top discords
    profile = discords(profile)

    # plot mp
    figures = visualize(profile)

    return (profile, figures)


def analyze(ts, query=None, windows=None, sample_pct=1.0, threshold=0.98, n_jobs=1,
            preprocessing_kwargs = None):
    """
    Runs an appropriate workflow based on the parameters passed in. The goal
    of this function is to compute all fundamental algorithms on the provided
    time series data. For now the following is computed:

    1. Matrix Profile - exact or approximate based on sample_pct given that a
       window is provided. By default is the exact algorithm.
    2. Top Motifs - The top 3 motifs are found.
    3. Top Discords - The top 3 discords are found.
    4. Plot MP, Motifs and Discords

    When a window is not provided or more than a single window is provided,
    the PMP is computed:

    1. Compute UPPER window when no window(s) is provided
    2. Compute PMP for all windows
    3. Top Motifs
    4. Top Discords
    5. Plot PMP, motifs and discords.

    Parameters
    ----------
    ts : array_like
        The time series to analyze.
    query : array_like, Optional
        The query to analyze. Note that when computing the PMP the query is
        ignored!
    windows : int or array_like, Optional
        The window(s) to compute the MatrixProfile. Note that it may be an int
        for a single matrix profile computation or an array of ints for
        computing the pan matrix profile.
    sample_pct : float, default = 1
        A float between 0 and 1 representing how many samples to compute for
        the MP or PMP. When it is 1, the exact algorithm is used.
    threshold : float, Default 0.98
        The correlation coefficient used as the threshold. It should be between
        0 and 1. This is used to compute the upper window size when no
        window(s) is given.
    n_jobs : int, Default = 1
        Number of cpu cores to use.
    preprocessing_kwargs : dict, default = None
        A dictionary object to sets parameters for preprocess function.
        A valid preprocessing_kwargs should have the following structure:

        >>> {
        >>>     'window': The window size to compute the mean/median/minimum/maximum value,
        >>>     'method': A string indicating the data imputation method, which should be
        >>>               'mean', 'median', 'min' or 'max',
        >>>     'direction': A string indicating the data imputation direction, which should be
        >>>                 'forward', 'fwd', 'f', 'backward', 'bwd', 'b'. If the direction is
        >>>                 forward, we use previous data for imputation; if the direction is
        >>>                 backward, we use subsequent data for imputation.,
        >>>     'add_noise': A boolean value indicating whether noise needs to be added into the
        >>>                 time series
        >>> }

        To disable preprocessing procedure, set the preprocessing_kwargs to
        None/False/""/{}.

    Returns
    -------
    tuple : (profile, figures)
        The appropriate PMP or MP profile object and associated figures.

    """
    result = None

    # preprocess the time series
    preprocessing_kwargs = validate_preprocess_kwargs(preprocessing_kwargs)
    if preprocessing_kwargs:
        ts = preprocess(ts,
                        window=preprocessing_kwargs['window'],
                        impute_method=preprocessing_kwargs['impute_method'],
                        impute_direction=preprocessing_kwargs['impute_direction'],
                        add_noise=preprocessing_kwargs['add_noise'])

    # determine proper number of jobs
    n_jobs = core.valid_n_jobs(n_jobs)

    # determine what algorithm to use based on params
    no_window = isinstance(windows, type(None))
    many_windows = core.is_array_like(windows) and len(windows) > 1
    single_window = isinstance(windows, int) or \
                    (core.is_array_like(windows) and len(windows) == 1)
    is_exact = sample_pct >= 1
    is_approx = sample_pct > 0 and sample_pct < 1

    # use PMP with no window provided
    if no_window or many_windows:
        result = analyze_pmp(ts, query, sample_pct, threshold, windows=windows, n_jobs=n_jobs)
    elif single_window and is_exact:
        result = analyze_mp_exact(ts, query, windows, n_jobs=n_jobs)
    elif single_window and is_approx:
        result = analyze_mp_approximate(ts, query, windows, sample_pct, n_jobs=n_jobs)
    else:
        raise RuntimeError('Param combination resulted in an uknown operation')

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import numpy as np

from matrixprofile import core
from matrixprofile.algorithms.mpdist import mpdist

def compute_dist(args):
    """
    Helper function to parallelize pairwise distance calculation.

    Parameters
    ----------
    args : tuple
        The arguments to pass to the mpdist calculation.
    
    Returns
    -------
    values : tuple
        The kth index and distance.
    """
    k = args[0]
    distance = mpdist(args[1], args[2], args[3], threshold=args[4])

    return (k, distance)


def pairwise_dist(X, window_size, threshold=0.05, n_jobs=1):
    """
    Utility function to compute all pairwise distances between the timeseries
    using MPDist. 
    
    Note
    ----
    scipy.spatial.distance.pdist cannot be used because they
    do not allow for jagged arrays, however their code was used as a reference
    in creating this function.
    https://github.com/scipy/scipy/blob/master/scipy/spatial/distance.py#L2039

    Parameters
    ----------
    X : array_like
        An array_like object containing time series to compute distances for.
    window_size : int
        The window size to use in computing the MPDist.
    threshold : float
        The threshold used to compute MPDist.
    n_jobs : int
        Number of CPU cores to use during computation.
    
    Returns
    -------
    Y : np.ndarray
        Returns a condensed distance matrix Y.  For
        each :math:`i` and :math:`j` (where :math:`i<j<m`),where m is the 
        number of original observations. The metric ``dist(u=X[i], v=X[j])``
        is computed and stored in entry ``ij``.
    """
    if not core.is_array_like(X):
        raise ValueError('X must be array_like!')
    
    # identify shape based on iterable or np.ndarray.shape
    m = 0
    
    if isinstance(X, np.ndarray) and len(X.shape) == 2:
        m = X.shape[0]
    else:
        m = len(X)
    
    dm = np.empty((m * (m - 1)) // 2, dtype=np.double)
    k = 0

    if n_jobs == 1:
        for i in range(0, m - 1):
            for j in range(i + 1, m):
                dm[k] = mpdist(X[i], X[j], window_size, threshold=threshold, 
                            n_jobs=n_jobs)
                k = k + 1
    else:
        args = []
        for i in range(0, m - 1):
            for j in range(i + 1, m):
                args.append((k, X[i], X[j], window_size, threshold))
                k = k + 1
        
        with core.mp_pool()(n_jobs) as pool:
            results = pool.map(compute_dist, args)
        
        # put results in the matrix
        for result in results:
            dm[result[0]] = result[1]
    
    return dm



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import numpy as np

from matrixprofile import core


def statistics(ts, window_size):
    """
    Compute global and moving statistics for the provided 1D time
    series. The statistics computed include the min, max, mean, std. and median
    over the window specified and globally.

    Parameters
    ----------
    ts : array_like
        The time series.
    window_size: int
        The size of the window to compute moving statistics over.

    Returns
    -------
    dict : statistics
        The global and rolling window statistics.
        
        >>> {
        >>>     ts: the original time series,
        >>>     min: the global minimum,
        >>>     max: the global maximum,
        >>>     mean: the global mean,
        >>>     std: the global standard deviation,
        >>>     median: the global median,
        >>>     moving_min: the moving minimum,
        >>>     moving_max: the moving maximum,
        >>>     moving_mean: the moving mean,
        >>>     moving_std: the moving standard deviation,
        >>>     moving_median: the moving median,
        >>>     window_size: the window size provided,
        >>>     class: Statistics
        >>> }

    Raises
    ------
    ValueError
        If window_size is not an int.
        If window_size > len(ts)
        If ts is not a list or np.array.
        If ts is not 1D.

    """
    if not core.is_array_like(ts):
        raise ValueError('ts must be array like')	

    if not core.is_one_dimensional(ts):
        raise ValueError('The time series must be 1D')

    if not isinstance(window_size, int):
        raise ValueError('Expecting int for window_size')

    if window_size > len(ts):
        raise ValueError('Window size cannot be greater than len(ts)')

    if window_size < 3:
        raise ValueError('Window size cannot be less than 3')

    moving_mu, moving_sigma = core.moving_avg_std(ts, window_size)
    rolling_ts = core.rolling_window(ts, window_size)

    return {
        'ts': ts,
        'min': np.min(ts),
        'max': np.max(ts),
        'mean': np.mean(ts),
        'std': np.std(ts),
        'median': np.median(ts),
        'moving_min': np.min(rolling_ts, axis=1),
        'moving_max': np.max(rolling_ts, axis=1),
        'moving_mean': moving_mu,
        'moving_std': moving_sigma,
        'moving_median': np.median(rolling_ts, axis=1),
        'window_size': window_size,
        'class': 'Statistics'
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

# handle Python 2/3 Iterable import
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import os

import numpy as np

from matplotlib import pyplot as plt
from matplotlib.lines import Line2D

from matrixprofile import core


def __combine(a, b):
    """
    Helper function to combine lists or a list and an object.
    """
    output = []
    if isinstance(a, list) and isinstance(b, list):
        output = a + b
    elif isinstance(a, list):
        output = a
        output.append(b)
    elif isinstance(b, list):
        output = b
        output.append(a)

    return output


def is_visualizable(obj):
    """
    Helper function to determine if the passed in object can be visualized or
    not based on the data structure.

    Parameters
    ----------
    obj : Object
        The object to test.

    Returns
    -------
    list : figures
        A list of matplotlib figures.

    """
    return core.is_mp_obj(obj) or core.is_pmp_obj(obj) or core.is_stats_obj(obj)


def visualize(profile):
    """
    Automatically creates plots for the provided data structure. In some cases
    many plots are created. For example, when a MatrixProfile is passed with
    corresponding motifs and discords, the matrix profile, discords and motifs
    will be plotted.

    Parameters
    ----------
    profile : dict_like
        A MatrixProfile, Pan-MatrixProfile or Statistics data structure.

    Returns
    -------
    list : figures
        A list of matplotlib figures.

    """
    figures = []

    if not is_visualizable(profile):
        raise ValueError('MatrixProfile, Pan-MatrixProfile or Statistics data structure expected!')

    # plot MP
    if core.is_mp_obj(profile):
        figures = __combine(figures, plot_mp(profile))

        if 'cmp' in profile and len(profile['cmp']) > 0:
            figures = __combine(figures, plot_cmp_mp(profile))

        if 'av' in profile and len(profile['av']) > 0:
            figures = __combine(figures, plot_av_mp(profile))

        if 'motifs' in profile and len(profile['motifs']) > 0:
            figures = __combine(figures, plot_motifs_mp(profile))

        if 'discords' in profile and len(profile['discords']) > 0:
            figures = __combine(figures, plot_discords_mp(profile))

    # plot PMP
    if core.is_pmp_obj(profile):
        figures = __combine(figures, plot_pmp(profile))

        if 'motifs' in profile and len(profile['motifs']) > 0:
            figures = __combine(figures, plot_motifs_pmp(profile))

        if 'discords' in profile and len(profile['discords']) > 0:
            figures = __combine(figures, plot_discords_pmp(profile))

    # plot stats
    if core.is_stats_obj(profile):
        figures = __combine(figures, plot_stats(profile))


    return figures


def plot_stats(profile):
    """
    Plots the given Statistics data structure provided.

    Parameters
    ----------
    profile : dict
        The dict structure from a Statistics algorithm.

    Returns
    -------
    matplotlib.Figure : figure
        The matplotlib figure object.

    """
    fig, ax = plt.subplots(2, 1, figsize=(15, 7))
    ts = profile.get('ts')
    ax[0].plot(ts, label='Time Series', c='black')

    for k, v in profile.items():
        if k.startswith('moving'):
            ax[1].plot(v, label=k)

    fig.legend(loc="upper right", bbox_to_anchor=(1.11, 0.97))
    fig.tight_layout()

    return fig


def plot_pmp(profile):
    """
    Plots the given Pan-MatrixProfile data structure provided.

    Parameters
    ----------
    profile : dict
        The dict structure from a PMP algorithm.

    Returns
    -------
    matplotlib.Figure : figure
        The matplotlib figure object.

    """
    pmp = profile.get('pmp', None)
    fig, ax = plt.subplots(1, 1, figsize=(10, 10))
    depth = 256
    test = np.ceil(pmp * depth) / depth
    test[test > 1] = 1

    ax.imshow(test, interpolation=None, aspect='auto')
    ax.invert_yaxis()
    ax.set_title('Pan-MatrixProfile')
    ax.set_xlabel('Profile Index')
    ax.set_ylabel('Window Size')
    ax.set_aspect(1, 'box')

    fig.tight_layout()

    return fig


def plot_mp(profile):
    """
    Plots the matrix profile given the appropriate data structure.

    Parameters
    ----------
    profile : dict_like
        The matrix profile to plot.

    Returns
    -------
    matplotlib.Figure : figure
        The matplotlib figure object.

    """
    plot_count = 0
    data = profile.get('data', None)
    ts = None
    query = None
    if data:
        ts = data.get('ts', None)
        query = data.get('query', None)

    mp = profile.get('mp', None)
    lmp = profile.get('lmp', None)
    rmp = profile.get('rmp', None)

    for val in [ts, query, mp, lmp, rmp]:
        if core.is_array_like(val):
            plot_count += 1

    if plot_count < 1:
        raise ValueError("Object passed has nothing to plot!")

    w = profile.get('w', None)
    if not isinstance(w, int):
        raise ValueError("Expecting window size!")

    current = 0

    fig, axes = plt.subplots(plot_count, 1, sharex=True, figsize=(15, 7))

    if not isinstance(axes, Iterable):
        axes = [axes,]

    # plot the original ts
    if core.is_array_like(ts):
        axes[current].plot(np.arange(len(ts)), ts)
        axes[current].set_ylabel('Data')
        current += 1

    # plot the original query
    if core.is_array_like(query):
        axes[current].plot(np.arange(len(query)), query)
        axes[current].set_ylabel('Query')
        current += 1

    # plot matrix profile
    if core.is_array_like(mp):
        mp_adj = np.append(mp, np.zeros(w - 1) + np.nan)
        axes[current].plot(np.arange(len(mp_adj)), mp_adj)
        axes[current].set_ylabel('Matrix Profile')
        axes[current].set_title('Window Size {}'.format(w))
        current += 1

    # plot left matrix profile
    if core.is_array_like(lmp):
        mp_adj = np.append(lmp, np.zeros(w - 1) + np.nan)
        axes[current].plot(np.arange(len(mp_adj)), mp_adj)
        axes[current].set_ylabel('Left Matrix Profile')
        axes[current].set_title('Window Size {}'.format(w))
        current += 1

    # plot left matrix profile
    if core.is_array_like(rmp):
        mp_adj = np.append(rmp, np.zeros(w - 1) + np.nan)
        axes[current].plot(np.arange(len(mp_adj)), mp_adj)
        axes[current].set_ylabel('Right Matrix Profile')
        axes[current].set_title('Window Size {}'.format(w))
        current += 1

    fig.tight_layout()

    return fig


def plot_cmp_mp(profile):
    """
    Plot corrected matrix profile for a MatrixProfile data structure.

    Parameters
    ----------
    profile : dict_like
        The matrix profile object to plot.

    Returns
    -------
    matplotlib.Figure : figure
        The matplotlib figure object.

    """
    cmp = profile['cmp']
    w = profile['w']

    fig, ax = plt.subplots(1, 1, figsize=(15, 7))

    cmp_adj = np.append(cmp, np.zeros(w - 1) + np.nan)
    ax.plot(np.arange(len(cmp_adj)), cmp_adj)
    ax.set_ylabel('Corrected Matrix Profile')
    ax.set_title('Window Size {}'.format(w))

    fig.tight_layout()

    return fig


def plot_av_mp(profile):
    """
    Plot the annotation vector for a MatrixProfile data structure.

    Parameters
    ----------
    profile : dict_like
        The matrix profile object to plot.

    Returns
    -------
    matplotlib.Figure : figure
        The matplotlib figure object.

    """
    av = profile['av']
    w = profile['w']

    fig, ax = plt.subplots(1, 1, figsize=(15, 7))

    av_adj = np.append(av, np.zeros(w - 1) + np.nan)
    ax.plot(np.arange(len(av_adj)), av_adj)
    ax.set_ylabel('Annotation Vector')
    ax.set_title('Window Size {}'.format(w))

    fig.tight_layout()

    return fig


def plot_discords_mp(profile):
    """
    Plot discords for a MatrixProfile data structure.

    Parameters
    ----------
    profile : dict_like
        The matrix profile object to plot.

    Returns
    -------
    matplotlib.Figure : figure
        The matplotlib figure object.

    """
    mp = profile['mp']
    w = profile['w']
    discords = profile['discords']
    data = profile.get('data', None)
    if data:
        ts = data.get('ts', None)

    mp_adjusted = np.append(mp, np.full(w + 1, np.nan))

    fig, axes = plt.subplots(3, 1, sharex=True, figsize=(15, 7), gridspec_kw={'height_ratios': [25, 5, 25]})

    pos = axes[1].imshow([mp_adjusted,], aspect='auto', cmap='coolwarm')
    axes[1].set_yticks([])
    axes[0].plot(np.arange(len(ts)), ts)
    axes[0].set_ylabel('Data')

    axes[2].plot(np.arange(len(mp_adjusted)), mp_adjusted)
    axes[2].set_ylabel('Matrix Profile')
    axes[2].set_title('Window Size {}'.format(w))

    for idx in discords:
        axes[2].plot(idx, mp_adjusted[idx], c='r', marker='*', lw=0, markersize=10)

    fig.subplots_adjust(right=0.8)
    cbar_ax = fig.add_axes([1, 0.46, 0.01, 0.1])
    fig.colorbar(pos, orientation='vertical', cax=cbar_ax, use_gridspec=True)

    lines = [
        Line2D([0], [0], color='red', marker='*', lw=0),
        Line2D([0], [0], color='blue'),
    ]
    fig.legend(lines, ['Discord', 'MP'], bbox_to_anchor=(1.07, 0.44))

    fig.tight_layout()

    return fig


def plot_discords_pmp(profile):
    """
    Plot discords for the given Pan-MatrixProfile data structure.

    Parameters
    ----------
    profile : dict_like
        The pmp object to plot.

    Returns
    -------
    matplotlib.Figure : figure
        The matplotlib figure object.

    """
    discord_figures = []

    for discord in profile['discords']:
        mp_idx = discord[0]
        idx = discord[1]
        w = profile['windows'][mp_idx]

        data = profile.get('data', None)
        if data:
            ts = data.get('ts', None)

        mp_adjusted = profile['pmp'][mp_idx]
        # mp_adjusted = np.append(mp, np.full(w + 1, np.nan))

        fig, axes = plt.subplots(3, 1, sharex=True, figsize=(15, 7), gridspec_kw={'height_ratios': [25, 5, 25]})

        pos = axes[1].imshow([mp_adjusted,], aspect='auto', cmap='coolwarm')
        axes[1].set_yticks([])
        axes[0].plot(np.arange(len(ts)), ts)
        axes[0].set_ylabel('Data')

        axes[2].plot(np.arange(len(mp_adjusted)), mp_adjusted)
        axes[2].set_ylabel('Matrix Profile')

        # for idx in discords:
        axes[2].plot(idx, mp_adjusted[idx], c='r', marker='*', lw=0, markersize=10)
        axes[2].set_title('Window Size = {}'.format(w))

        fig.subplots_adjust(right=0.8)
        cbar_ax = fig.add_axes([1, 0.46, 0.01, 0.1])
        fig.colorbar(pos, orientation='vertical', cax=cbar_ax, use_gridspec=True)

        lines = [
            Line2D([0], [0], color='red', marker='*', lw=0),
            Line2D([0], [0], color='blue'),
        ]
        fig.legend(lines, ['Discord', 'MP'], bbox_to_anchor=(1.07, 0.44))

        fig.tight_layout()

        discord_figures.append(fig)

    return discord_figures


def plot_motifs_mp(profile):
    """
    Plot motifs given a MatrixProfile data structure.

    Parameters
    ----------
    profile : dict_like
        The matrix profile object to plot.

    Returns
    -------
    list : figures
        A list of matplotlib figure objects.

    """
    figures = []

    w = profile['w']
    motifs = profile['motifs']
    data = profile.get('data', None)
    if data:
        ts = data.get('ts', None)

    fig, axes = plt.subplots(len(motifs), 2, figsize=(15, 7), sharey='row', sharex='col')
    if len(motifs) == 1:
        axes = [axes,]

    pair_num = 1
    for ax_row, motif in zip(axes, motifs):
        first = True
        for ax, idx in zip(ax_row, motif['motifs']):
            subquery = ts[idx:idx + w]
            indices = np.arange(len(subquery))
            ax.plot(indices, subquery)
            ax.set_title('Index Start {}'.format(idx))
            if first:
                ax.set_ylabel('Motif {}'.format(pair_num))
                first = False

        pair_num += 1

    fig.tight_layout()
    figures.append(fig)

    fig, axes = plt.subplots(len(motifs), 1, figsize=(15, 7), sharey='row', sharex='col')
    if len(motifs) == 1:
        axes = [axes,]

    pair_num = 1
    for ax, motif in zip(axes, motifs):
        ax.plot(np.arange(len(ts)), ts)
        for idx in motif['motifs']:
            subquery = ts[idx:idx + w]
            indices = np.arange(idx, idx + w)
            ax.plot(indices, subquery, c='r')
            ax.set_ylabel('Motif {}'.format(pair_num))

        for idx in motif['neighbors']:
            subquery = ts[idx:idx + w]
            indices = np.arange(idx, idx + w)
            ax.plot(indices, subquery, c='black')

        pair_num += 1

    lines = [
        Line2D([0], [0], color='blue'),
        Line2D([0], [0], color='red'),
        Line2D([0], [0], color='black')
    ]
    fig.legend(lines, ['Data', 'Motif', 'Neighbor'], bbox_to_anchor=(1.08, 0.975))
    fig.tight_layout()

    figures.append(fig)

    return figures


def plot_motifs_pmp(profile):
    """
    Plot motifs given a Pan-MatrixProfile data structure.

    Parameters
    ----------
    profile : dict_like
        The pan matrix profile object to plot.

    Returns
    -------
    list : figures
        A list of matplotlib figure objects.

    """
    motif_figures = []
    motifs = profile.get('motifs')
    data = profile.get('data', None)
    windows = profile.get('windows', None)
    if data:
        ts = data.get('ts', None)

    fig, axes = plt.subplots(len(motifs), 2, figsize=(15, 7), sharey='row', sharex='col')
    if len(motifs) == 1:
        axes = [axes,]

    pair_num = 1
    for ax_row, motif in zip(axes, motifs):
        first = True
        for ax, motif_loc in zip(ax_row, motif['motifs']):
            w = windows[motif_loc[0]]
            idx = motif_loc[1]
            subquery = ts[idx:idx + w]
            indices = np.arange(len(subquery))
            ax.plot(indices, subquery)
            ax.set_title('Index Start {}, Window Size {}'.format(idx, w))
            if first:
                ax.set_ylabel('Motif {}'.format(pair_num))
                first = False

        pair_num += 1

    fig.tight_layout()
    motif_figures.append(fig)

    fig, axes = plt.subplots(len(motifs), 1, figsize=(15, 7), sharey='row', sharex='col')
    if len(motifs) == 1:
        axes = [axes,]

    pair_num = 1
    for ax, motif in zip(axes, motifs):
        ax.plot(np.arange(len(ts)), ts)
        for motif_loc in motif['motifs']:
            w = windows[motif_loc[0]]
            idx = motif_loc[1]
            subquery = ts[idx:idx + w]
            indices = np.arange(idx, idx + w)
            ax.plot(indices, subquery, c='r')
            ax.set_title('Window Size {}'.format(w))
            ax.set_ylabel('Motif {}'.format(pair_num))

        for neigh_loc in motif['neighbors']:
            w = windows[neigh_loc[0]]
            idx = neigh_loc[1]
            subquery = ts[idx:idx + w]
            indices = np.arange(idx, idx + w)
            ax.plot(indices, subquery, c='black')

        pair_num += 1

    lines = [
        Line2D([0], [0], color='blue'),
        Line2D([0], [0], color='red'),
        Line2D([0], [0], color='black')
    ]
    fig.legend(lines, ['Data', 'Motif', 'Neighbor'], bbox_to_anchor=(1.08, 0.975))
    fig.tight_layout()

    motif_figures.append(fig)

    return motif_figures


def plot_snippets(snippets, ts):
    """
    Plot snippets for the given Snippets data structure.
    
    Parameters
    ----------
    snippets : list
        A list of snippets as dictionary objects to plot.
    ts : array_like
        The time series.
        
    Returns
    -------
    list : figures
        A list of matplotlib figures.
        
    """
    figures = []
    
    for i in range(len(snippets)):
        snippet_id = str(i+1)
        snippet_start = snippets[i]['index']
        snippet_end = snippets[i]['index']+len(snippets[i]['snippet'])
        snippet_data = snippets[i]['snippet']
        
        # Create a plot for current snippets
        fig, ax = plt.subplots(1, 1, sharex=True, figsize=(15,5))
        ax.plot(ts)
        ax.set_title('Snippet-'+snippet_id, size=12)
        ax.set_ylabel('Data')
        flag = 1
    
        # Get intervals for the given neighboring snippet indices
        neighbors = snippets[i]['neighbors']
        intervals = []
        for i in range(len(neighbors)):
            if i == 0:
                intervals.append(neighbors[i])
            if i == len(neighbors)-1:
                intervals.append(neighbors[i])
                break
            if (neighbors[i+1] - neighbors[i]) != 1:
                intervals.append(neighbors[i])
                intervals.append(neighbors[i+1])
        step = 2
        intervals = [intervals[i:i+step] for i in range(0, len(intervals), step)]
    
        # Plot the neighboring snippets
        for interval in intervals:
            start = interval[0]
            end = interval[1]
            if flag:
                ax.plot(np.arange(start,end+1),ts[start:end+1], c = "orange"
                        ,label = "Subsequences Represented by Snippet-"+ snippet_id)
                flag = 0
            else:
                ax.plot(np.arange(start,end+1),ts[start:end+1],"orange")
    
        # Plot the snippet
        ax.plot(np.arange(snippet_start,snippet_end), snippet_data, c = "red"
                , label = 'Snippet-'+ snippet_id)
        
        plt.legend(loc="upper right")       
        fig.tight_layout()       
        figures.append(fig)
    
    return figures




from matrixprofile.algorithms.stomp import stomp
from matrixprofile.algorithms.mass2 import mass2
from matrixprofile.algorithms.mpdist import mpdist
from matrixprofile.algorithms.mpdist import mpdist_vector
from matrixprofile.algorithms.mpx import mpx
from matrixprofile.algorithms.scrimp import scrimp_plus_plus
from matrixprofile.algorithms.scrimp import prescrimp
from matrixprofile.algorithms.skimp import skimp
from matrixprofile.algorithms.skimp import maximum_subsequence
from matrixprofile.algorithms.pairwise_dist import pairwise_dist


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.statistics import statistics


def test_invalid_ts_not_1d():
    ts = np.array([[1, 1], [1, 1]])
    w = 2

    with pytest.raises(ValueError) as excinfo:
        statistics(ts, w)
        assert 'The time series must be 1D' in str(excinfo.value)


def test_invalid_ts_not_array():
    ts = None
    w = 2

    with pytest.raises(ValueError) as excinfo:
        statistics(ts, w)
        assert 'ts must be array like' in str(excinfo.value)


def test_invalid_window_size_not_int():
    ts = np.arange(10)
    w = 's'

    with pytest.raises(ValueError) as excinfo:
        statistics(ts, w)
        assert 'Expecting int for window_size' in str(excinfo.value)


def test_invalid_window_size_too_large():
    ts = np.arange(10)
    w = 11

    with pytest.raises(ValueError) as excinfo:
        statistics(ts, w)
        assert 'Window size cannot be greater than len(ts)' in str(excinfo.value)


def test_invalid_window_size_too_small():
    ts = np.arange(10)
    w = 2

    with pytest.raises(ValueError) as excinfo:
        statistics(ts, w)
        assert 'Window size cannot be less than 3' in str(excinfo.value)


def test_valid():
	ts = np.array([1, 3, 2, 4, 5, 1, 1, 1, 2, 4, 9, 7])
	w = 4
	ts_stats = statistics(ts, w)

	assert(ts_stats['min'] == 1)
	assert(ts_stats['max'] == 9)
	np.testing.assert_almost_equal(ts_stats['mean'], 3.3333333)
	np.testing.assert_almost_equal(ts_stats['std'], 2.494438257)
	assert(ts_stats['median'] == 2.5)
	np.testing.assert_almost_equal(ts_stats['moving_min'], np.array([1, 2, 1, 1, 1, 1, 1, 1, 2]))
	np.testing.assert_almost_equal(ts_stats['moving_max'], np.array([4, 5, 5, 5, 5, 2, 4, 9, 9]))
	np.testing.assert_almost_equal(ts_stats['moving_mean'], np.array([2.5, 3.5, 3.0, 2.75, 2.0, 1.25, 2.0, 4.0, 5.5]))
	np.testing.assert_almost_equal(ts_stats['moving_std'], np.array([1.11803399, 1.11803399, 1.58113883, 1.78535711, 1.73205081, 0.4330127, 1.22474487, 3.082207, 2.6925824]))
	np.testing.assert_almost_equal(ts_stats['moving_median'], np.array([2.5, 3.5, 3.0, 2.5, 1.0, 1.0, 1.5, 3.0, 5.5]))
	np.testing.assert_equal(ts_stats['ts'], ts)
	assert(ts_stats['window_size'] == w)
	assert(ts_stats['class'] == 'Statistics')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.hierarchical_clustering import (
    hierarchical_clusters
)

def test_hierarchical_clusters_valid_simple():
    np.random.seed(9999)
    ts = np.random.uniform(size=2**10)
    ts2 = np.random.uniform(size=2**10)
    ts3 = np.random.uniform(size=2**10)
    X = [
        ts,
        ts,
        ts2,
        ts2,
        ts3
    ]
    w = 2**6
    t = 2

    clusters = hierarchical_clusters(X, w, t)

    # evaluate keys
    expected_keys = set([
        'pairwise_distances',
        'linkage_matrix',
        'inconsistency_statistics',
        'assignments',
        'cophenet',
        'cophenet_distances',
        'class'
    ])
    actual_keys = set(clusters.keys())
    assert(expected_keys == actual_keys)
    assert(clusters['class'] == 'hclusters')

    # evaluate cluster assignments
    expected_assignments = np.array([1, 1, 2, 2, 3])
    np.testing.assert_equal(clusters['assignments'], expected_assignments)

    # evaluate cophenet score
    expected_cophenet = 0.9999870997174531
    np.testing.assert_almost_equal(clusters['cophenet'], expected_cophenet)

    # evaluate pairwise distances
    expected_distances = np.array([0, 8.2299501, 8.2299501, 8.29915377, 
    8.2299501, 8.2299501, 8.29915377, 0, 8.2558308, 8.2558308])
    np.testing.assert_almost_equal(
        clusters['pairwise_distances'], expected_distances)


def test_hierarchical_clusters_valid_simple_parallel():
    np.random.seed(9999)
    ts = np.random.uniform(size=2**10)
    ts2 = np.random.uniform(size=2**10)
    ts3 = np.random.uniform(size=2**10)
    X = [
        ts,
        ts,
        ts2,
        ts2,
        ts3
    ]
    w = 2**6
    t = 2

    clusters = hierarchical_clusters(X, w, t, n_jobs=2)

    # evaluate keys
    expected_keys = set([
        'pairwise_distances',
        'linkage_matrix',
        'inconsistency_statistics',
        'assignments',
        'cophenet',
        'cophenet_distances',
        'class'
    ])
    actual_keys = set(clusters.keys())
    assert(expected_keys == actual_keys)
    assert(clusters['class'] == 'hclusters')

    # evaluate cluster assignments
    expected_assignments = np.array([1, 1, 2, 2, 3])
    np.testing.assert_equal(clusters['assignments'], expected_assignments)

    # evaluate cophenet score
    expected_cophenet = 0.9999870997174531
    np.testing.assert_almost_equal(clusters['cophenet'], expected_cophenet)

    # evaluate pairwise distances
    expected_distances = np.array([0, 8.2299501, 8.2299501, 8.29915377, 
    8.2299501, 8.2299501, 8.29915377, 0, 8.2558308, 8.2558308])
    np.testing.assert_almost_equal(
        clusters['pairwise_distances'], expected_distances)


def test_hierarchical_clusters_invalid_params():
    np.random.seed(9999)
    ts = np.random.uniform(size=2**10)
    ts2 = np.random.uniform(size=2**10)
    ts3 = np.random.uniform(size=2**10)
    X = [
        ts,
        ts,
        ts2,
        ts2,
        ts3
    ]
    w = 2**6
    t = 2

    # invalid X
    with pytest.raises(ValueError) as excinfo:
        clusters = hierarchical_clusters('', w, t)
        assert('X must be array_like!' == str(excinfo.value))

    # invalid t
    with pytest.raises(ValueError) as excinfo:
        clusters = hierarchical_clusters(X, w, '')
        assert('t must be a scalar (int or float)' == str(excinfo.value))
    
    # invalid threshold 0
    error = 'threshold must be a float greater than 0 and less than 1'
    with pytest.raises(ValueError) as excinfo:
        clusters = hierarchical_clusters(X, w, t, threshold=0)
        assert(error == str(excinfo.value))
    
    # invalid threshold 1
    with pytest.raises(ValueError) as excinfo:
        clusters = hierarchical_clusters(X, w, t, threshold=1)
        assert(error == str(excinfo.value))

    # invalid threshold not numeric
    with pytest.raises(ValueError) as excinfo:
        clusters = hierarchical_clusters(X, w, t, threshold='')
        assert(error == str(excinfo.value))

    # invalid depth < 1
    error = 'depth must be an integer greater than 0'
    with pytest.raises(ValueError) as excinfo:
        clusters = hierarchical_clusters(X, w, t, depth=0)
        assert(error == str(excinfo.value))

    # invalid depth not int
    with pytest.raises(ValueError) as excinfo:
        clusters = hierarchical_clusters(X, w, t, depth='')
        assert(error == str(excinfo.value))

    # invalid method
    with pytest.raises(ValueError) as excinfo:
        clusters = hierarchical_clusters(X, w, t, method='')
        assert('method may be only one of' in str(excinfo.value))

    # invalid criterion
    with pytest.raises(ValueError) as excinfo:
        clusters = hierarchical_clusters(X, w, t, criterion='')
        assert('criterion may be only one of' in str(excinfo.value))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

# Third-party imports
import numpy as np

# Project imports
from matrixprofile import core


def validate_preprocess_kwargs(preprocessing_kwargs):
    """
    Tests the arguments of preprocess function and raises errors for invalid arguments.

    Parameters
    ----------
    preprocessing_kwargs : dict-like or None or False
        A dictionary object to store keyword arguments for the preprocess function.
        It can also be None/False/{}/"".

    Returns
    -------
    valid_kwargs : dict-like or None
        The valid keyword arguments for the preprocess function.
        Returns None if the input preprocessing_kwargs is None/False/{}/"".

    Raises
    ------
    ValueError
        If preprocessing_kwargs is not dict-like or None.
        If gets invalid key(s) for preprocessing_kwargs.
        If gets invalid value(s) for preprocessing_kwargs['window'], preprocessing_kwargs['impute_method']
        preprocessing_kwargs['impute_direction'] and preprocessing_kwargs['add_noise'].

    """
    if preprocessing_kwargs:

        valid_preprocessing_kwargs_keys = {'window', 'impute_method', 'impute_direction', 'add_noise'}

        if not isinstance(preprocessing_kwargs,dict):
            raise ValueError("The parameter 'preprocessing_kwargs' is not dict like!")

        elif set(preprocessing_kwargs.keys()).issubset(valid_preprocessing_kwargs_keys):
            window = 4
            impute_method = 'mean'
            impute_direction = 'forward'
            add_noise = True
            methods = ['mean', 'median', 'min', 'max']
            directions = ['forward', 'fwd', 'f', 'backward', 'bwd', 'b']

            if 'window' in preprocessing_kwargs.keys():
                if not isinstance(preprocessing_kwargs['window'],int):
                    raise ValueError("The value for preprocessing_kwargs['window'] is not an integer!")
                window = preprocessing_kwargs['window']


            if 'impute_method' in preprocessing_kwargs.keys():
                if preprocessing_kwargs['impute_method'] not in methods:
                    raise ValueError('invalid imputation method! valid include options: ' + ', '.join(methods))
                impute_method = preprocessing_kwargs['impute_method']

            if 'impute_direction' in preprocessing_kwargs.keys():
                if preprocessing_kwargs['impute_direction'] not in directions:
                    raise ValueError('invalid imputation direction! valid include options: ' + ', '.join(directions))
                impute_direction = preprocessing_kwargs['impute_direction']

            if 'add_noise' in preprocessing_kwargs.keys():
                if not isinstance(preprocessing_kwargs['add_noise'],bool):
                    raise ValueError("The value for preprocessing_kwargs['add_noise'] is not a boolean value!")
                add_noise = preprocessing_kwargs['add_noise']

            valid_kwargs =  { 'window': window,
                              'impute_method': impute_method,
                              'impute_direction': impute_direction,
                              'add_noise': add_noise }
        else:
            raise ValueError('invalid key(s) for preprocessing_kwargs! '
                             'valid key(s) should include '+ str(valid_preprocessing_kwargs_keys))
    else:
        valid_kwargs = None

    return valid_kwargs


def is_subsequence_constant(subsequence):
    """
    Determines whether the given time series subsequence is an array of constants.

    Parameters
    ----------
    subsequence : array_like
        The time series subsequence to analyze.

    Returns
    -------
    is_constant : bool
        A boolean value indicating whether the given subsequence is an array of constants.

    """
    if not core.is_array_like(subsequence):
        raise ValueError('subsequence is not array like!')

    temp = core.to_np_array(subsequence)
    is_constant = np.all(temp == temp[0])

    return is_constant


def add_noise_to_series(series):
    """
    Adds noise to the given time series.

    Parameters
    ----------
    series : array_like
        The time series subsequence to be added noise.

    Returns
    -------
    temp : array_like
        The time series subsequence after being added noise.

    """
    if not core.is_array_like(series):
        raise ValueError('series is not array like!')

    temp = np.copy(core.to_np_array(series))
    noise = np.random.uniform(0, 0.0000009, size=len(temp))
    temp = temp + noise

    return temp


def impute_missing(ts, window, method='mean', direction='forward'):
    """
    Imputes missing data in time series.

    Parameters
    ----------
    ts : array_like
        The time series to be handled.
    window : int
        The window size to compute the mean/median/minimum value/maximum
        value.
    method : string, Default = 'mean'
        A string indicating the data imputation method, which should be
        'mean', 'median', 'min' or 'max'.
    direction : string, Default = 'forward'
        A string indicating the data imputation direction, which should be
        'forward', 'fwd', 'f', 'backward', 'bwd', 'b'. If the direction is
        forward, we use previous data for imputation; if the direction is
        backward, we use subsequent data for imputation.

    Returns
    -------
    temp : array_like
        The time series after being imputed missing data.

    """
    method_map = {
        'mean': np.mean,
        'median': np.median,
        'min': np.min,
        'max': np.max
    }

    directions = ['forward', 'fwd', 'f', 'backward', 'bwd', 'b']

    if not core.is_array_like(ts):
        raise ValueError('ts is not array like!')

    if method not in method_map:
        raise ValueError('invalid imputation method! valid include options: {}'.format(', '.join(method_map.keys())))

    if direction not in directions:
        raise ValueError('invalid imputation direction! valid include options: ' + ', '.join(directions))

    if not isinstance(window, int):
        raise ValueError("window is not an integer!")

    temp = np.copy(core.to_np_array(ts))
    nan_infs = core.nan_inf_indices(temp)
    func = method_map[method]

    # Deal with missing data at the beginning and end of time series
    if np.isnan(temp[0]) or np.isinf(temp[0]):
        temp[0] = temp[~nan_infs][0]
        nan_infs = core.nan_inf_indices(temp)

    if np.isnan(temp[-1]) or np.isinf(temp[-1]):
        temp[-1] = temp[~nan_infs][-1]
        nan_infs = core.nan_inf_indices(temp)

    index_order = None

    if direction.startswith('f'):
        # Use previous data for imputation / fills in data in a forward direction
        index_order = range(len(temp) - window + 1)
    elif direction.startswith('b'):
        # Use subsequent data for imputation / fills in data in a backward direction
        index_order = range(len(temp) - window + 1, 0, -1)

    for index in index_order:
        start = index
        end = index + window
        has_missing = np.any(nan_infs[index:index + window])

        if has_missing:
            subseq = temp[start:end]
            nan_infs_subseq = nan_infs[start:end]
            stat = func(temp[start:end][~nan_infs_subseq])
            temp[start:end][nan_infs_subseq] = stat
            # Update nan_infs after array 'temp' is changed
            nan_infs = core.nan_inf_indices(temp)

    return temp


def preprocess(ts, window, impute_method='mean', impute_direction='forward', add_noise=True):
    """
    Preprocesses the given time series by adding noise and imputing missing data.

    Parameters
    ----------
    ts : array_like
        The time series to be preprocessed.
    window : int
        The window size to compute the mean/median/minimum value/maximum
        value.
    method : string, Default = 'mean'
        A string indicating the data imputation method, which should be
        'mean', 'median', 'min' or 'max'.
    direction : string, Default = 'forward'
        A string indicating the data imputation direction, which should be
        'forward', 'fwd', 'f', 'backward', 'bwd', 'b'. If the direction is
        forward, we use previous data for imputation; if the direction is
        backward, we use subsequent data for imputation.
    add_noise : bool, Default = True
        A boolean value indicating whether noise needs to be added into the time series.

    Returns
    -------
    temp : array_like
        The time series after being preprocessed.

    """
    if not core.is_array_like(ts):
        raise ValueError('ts is not array like!')

    temp = np.copy(core.to_np_array(ts))

    # impute missing
    temp = impute_missing(temp, window, method=impute_method, direction=impute_direction)

    # handle constant values
    if add_noise:
        for index in range(len(temp) - window + 1):
            start = index
            end = index + window
            subseq = temp[start:end]

            if is_subsequence_constant(subseq):
                temp[start:end] = add_noise_to_series(subseq)

    return temp


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

"""Tests for `mass_ts` package."""

import os

import pytest

import numpy as np

from matrixprofile import cycore
import matrixprofile

MODULE_PATH = matrixprofile.__path__[0]


def test_moving_avg_std():
    a = np.array([1, 2, 3, 4, 5, 6], dtype='d')
    mu, std = cycore.moving_avg_std(a, 3)
    mu_desired = np.array([2., 3., 4., 5.])
    std_desired = np.array([0.81649658, 0.81649658, 0.81649658, 0.81649658])

    np.testing.assert_almost_equal(mu, mu_desired)
    np.testing.assert_almost_equal(std, std_desired)


def test_it_should_not_produce_nan_values_when_std_is_almost_zero():
    a = np.array([10.1, 10.1, 10.1, 10.1, 10.1, 10.1, 10.1], dtype='d')
    mu, std = cycore.moving_avg_std(a, 3)
    mu_muinvn, std_muinvn = cycore.muinvn(a, 3)

    mu_desired = np.array([10.1, 10.1, 10.1, 10.1, 10.1])
    std_desired = np.array([0, 0, 0, 0, 0])

    np.testing.assert_almost_equal(mu, mu_desired)
    np.testing.assert_almost_equal(std, std_desired)

    np.testing.assert_almost_equal(mu_muinvn, mu_desired)
    np.testing.assert_almost_equal(std_muinvn, std_desired)


def test_moving_muinvn():
    a = np.array([1, 2, 3, 4, 5, 6], dtype='d')
    mu, std = cycore.muinvn(a, 3)
    mu_desired = np.array([2., 3., 4., 5.])
    std_desired = np.array([0.7071068, 0.7071068, 0.7071068, 0.7071068])

    np.testing.assert_almost_equal(mu, mu_desired)
    np.testing.assert_almost_equal(std, std_desired)


def test_muinvn_vs_matlab():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    w = 32

    ml_mu = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'muinvn_mua.txt'))
    ml_std = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'muinvn_stda.txt'))

    mu, std = cycore.muinvn(ts, w)

    np.testing.assert_almost_equal(ml_mu, mu, decimal=4)
    np.testing.assert_almost_equal(ml_std, std, decimal=4)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.top_k_discords import top_k_discords


def test_mp_all_same():
	profile = {
		'mp': np.ones(10),
		'ez': 2,
        'w': 4,
        'class': 'MatrixProfile'
	}

	discords = top_k_discords(profile)['discords']
	desired = np.array([9, 6, 3])
	np.testing.assert_almost_equal(discords, desired)


def test_discords_no_exclusion():
	profile = {
		'mp': np.array([1, 2, 3, 4]),
		'w': 4,
		'class': 'MatrixProfile'
	}
	desired = np.array([3, 2, 1])
	discords = top_k_discords(profile, k=3, exclusion_zone=0)['discords']
	np.testing.assert_almost_equal(discords, desired)


def test_discords_no_exclusion_all():
	profile = {
		'mp': np.array([1, 2, 3, 4]),
		'w': 4,
		'class': 'MatrixProfile'
	}
	desired = np.array([3, 2, 1, 0])
	discords = top_k_discords(profile, k=4, exclusion_zone=0)['discords']
	np.testing.assert_almost_equal(discords, desired)


def test_discords_exclude_one():
	profile = {
		'mp': np.array([1, 2, 3, 4]),
		'w': 4,
		'class': 'MatrixProfile'
	}
	desired = np.array([3, 1])
	discords = top_k_discords(profile, k=4, exclusion_zone=1)['discords']
	np.testing.assert_almost_equal(discords, desired)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import math

import numpy as np

from matrixprofile import core
from matrixprofile.algorithms.cympx import mpx_parallel as cympx_parallel
from matrixprofile.algorithms.cympx import mpx_ab_parallel as cympx_ab_parallel


def mpx(ts, w, query=None, cross_correlation=False, n_jobs=1):
    """
    The MPX algorithm computes the matrix profile without using the FFT.

    Parameters
    ----------
    ts : array_like
        The time series to compute the matrix profile for.
    w : int
        The window size.
    query : array_like
        Optionally a query series.
    cross_correlation : bool, Default=False
        Determine if cross_correlation distance should be returned. It defaults
        to Euclidean Distance.
    n_jobs : int, Default = 1
        Number of cpu cores to use.
    
    Returns
    -------
    dict : profile
        A MatrixProfile data structure.
        
        >>> {
        >>>     'mp': The matrix profile,
        >>>     'pi': The matrix profile 1NN indices,
        >>>     'rmp': The right matrix profile,
        >>>     'rpi': The right matrix profile 1NN indices,
        >>>     'lmp': The left matrix profile,
        >>>     'lpi': The left matrix profile 1NN indices,
        >>>     'metric': The distance metric computed for the mp,
        >>>     'w': The window size used to compute the matrix profile,
        >>>     'ez': The exclusion zone used,
        >>>     'join': Flag indicating if a similarity join was computed,
        >>>     'sample_pct': Percentage of samples used in computing the MP,
        >>>     'data': {
        >>>         'ts': Time series data,
        >>>         'query': Query data if supplied
        >>>     }
        >>>     'class': "MatrixProfile"
        >>>     'algorithm': "mpx"
        >>> }

    """
    ts = core.to_np_array(ts).astype('d')
    n_jobs = core.valid_n_jobs(n_jobs)
    is_join = False

    if core.is_array_like(query):
        query = core.to_np_array(query).astype('d')
        is_join = True
        mp, mpi, mpb, mpib = cympx_ab_parallel(ts, query, w, 
            int(cross_correlation), n_jobs)
    else:
        mp, mpi = cympx_parallel(ts, w, int(cross_correlation), n_jobs)

    mp = np.asarray(mp)
    mpi = np.asarray(mpi)
    distance_metric = 'euclidean'
    if cross_correlation:
        distance_metric = 'cross_correlation'

    return {
        'mp': mp,
        'pi': mpi,
        'rmp': None,
        'rpi': None,
        'lmp': None,
        'lpi': None,
        'metric': distance_metric,
        'w': w,
        'ez': int(np.ceil(w / 4.0)) if is_join else 0,
        'join': is_join,
        'sample_pct': 1,
        'data': {
            'ts': ts,
            'query': query
        },
        'class': 'MatrixProfile',
        'algorithm': 'mpx'
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.snippets import snippets

import matrixprofile

MODULE_PATH = matrixprofile.__path__[0]

def test_snippets():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    w = 32
    snippet_size = 64

    result = snippets(ts, snippet_size, window_size=w)
    assert(result[0]['index'] == 384)
    assert(result[1]['index'] == 704)
    assert(sum(result[0]['neighbors']) == 191408)
    assert(sum(result[1]['neighbors']) == 190967)

    # test inferred window size of snippet size / 2
    result = snippets(ts, snippet_size)
    assert(result[0]['index'] == 384)
    assert(result[1]['index'] == 704)
    assert(sum(result[0]['neighbors']) == 191408)
    assert(sum(result[1]['neighbors']) == 190967)

    snippet_size = 128
    result = snippets(ts, snippet_size, window_size=w)
    assert(result[0]['index'] == 384)
    assert(result[1]['index'] == 640)
    assert(sum(result[0]['neighbors']) == 227661)
    assert(sum(result[1]['neighbors']) == 154714)

    snippet_size = 8
    result = snippets(ts, snippet_size, window_size=snippet_size / 2)
    assert(result[0]['index'] == 72)
    assert(result[1]['index'] == 784)
    assert(sum(result[0]['neighbors']) == 149499)
    assert(sum(result[1]['neighbors']) == 232876)

def test_invalid_snippet_size():
    ts = np.arange(100)
    ss = 2

    error = 'snippet_size must be an integer >= 4'
    with pytest.raises(ValueError) as excinfo:
        snippets(ts, ss)
        assert(error == str(excinfo.value))

    with pytest.raises(ValueError) as excinfo:
        snippets(ts, '232')
        assert(error == str(excinfo.value))


def test_invalid_snippet_size_and_ts():
    ts = np.arange(100)
    ss = 75

    error = 'Time series is too short relative to snippet length'
    with pytest.raises(ValueError) as excinfo:
        snippets(ts, ss)
        assert(error == str(excinfo.value))


def test_window_size_greater_snippet_size():
    ts = np.arange(100)
    ss = 25
    w = 30

    error = 'window_size must be smaller than snippet_size'
    with pytest.raises(ValueError) as excinfo:
        snippets(ts, ss, window_size=w)
        assert(error == str(excinfo.value))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import numpy as np

from matrixprofile import core


def idealized_arc_curve(width, index):
    """
    Returns the value at x for the parabola of width n and height n / 2.
    Formula taken from https://www.desmos.com/calculator/awtnrxh6rk.

    Parameters
    ----------
    width : int
        Length of the time series to calculate the parabola for.
    index : int
        location to compute the parabola value at.

    Returns
    -------
    float : y
        The value at index for the parabola.

    """
    height = width / 2
    c = width / 2
    b = height
    a = height / (width / 2) ** 2
    y = -(a * (index - c) ** 2) + b

    return y


def fluss(profile):
    """
    Computes the corrected arc curve (CAC) for the MatrixProfile index. This
    algorithm is provides Fast Low-cost Unipotent Semantic Segmantation.

    Parameters
    ----------
    profile : dict
        Data structure from a MatrixProfile algorithm.

    Returns
    -------
    array_like : corrected_arc_curve
        The corrected arc curve for the profile.

    """
    if not core.is_mp_obj(profile):
        raise ValueError('profile must be a MatrixProfile structure')

    mpi = profile.get('pi')
    w = profile.get('w')

    n = len(mpi)
    nnmark = np.zeros(n)

    # find the number of additional arcs starting to cross over each index
    for i in range(n):
        mpi_val = mpi[i]
        small = int(min(i, mpi_val))
        large = int(max(i, mpi_val))
        nnmark[small + 1] = nnmark[small + 1] + 1
        nnmark[large] = nnmark[large] - 1

    # cumulatively sum all crossing arcs at each index
    cross_count = np.cumsum(nnmark)

    # compute ideal arc curve for all indices
    idealized = np.apply_along_axis(lambda i: idealized_arc_curve(n, i), 0, np.arange(0, n))
    idealized = cross_count / idealized

    # correct the arc curve so that it is between 0 and 1
    idealized[idealized > 1] = 1
    corrected_arc_curve = idealized

    # correct the head and tail with the window size
    corrected_arc_curve[:w] = 1
    corrected_arc_curve[-w:] = 1

    return corrected_arc_curve


def extract_regimes(profile, num_regimes=3):
    """
    Given a MatrixProfile, compute the corrected arc curve and extract
    the desired number of regimes. Regimes are computed with an exclusion
    zone of 5 * window size per the authors. 

    The author states:
        This exclusion zone is based on an assumption that regimes will have
        multiple repetitions; FLUSS is not able to segment single gesture 
        patterns.

    Parameters
    ----------
    profile : dict
        Data structure from a MatrixProfile algorithm.
    num_regimes : int
        The desired number of regimes to find.

    Returns
    -------
    dict : profile
        The original MatrixProfile object with additional keys containing.

        >>> {
        >>> 	'cac': The corrected arc curve
        >>> 	'cac_ez': The exclusion zone used
        >>> 	'regimes': Array of starting indices indicating a regime.
        >>> }

    """
    if not core.is_mp_obj(profile):
        raise ValueError('profile must be a MatrixProfile structure')

    cac = profile.get('cac')
    window_size = profile.get('w')
    ez = window_size * 5

    # compute the CAC if needed
    if isinstance(cac, type(None)):
        cac = fluss(profile)
        profile['cac'] = cac

    regimes = []
    tmp = np.copy(cac)
    n = len(tmp)
    
    for _ in range(num_regimes):
        min_index = np.argmin(tmp)
        regimes.append(min_index)
        
        # apply exclusion zone
        ez_start = np.max([0, min_index - ez])
        ez_end = np.min([n, min_index + ez])
        tmp[ez_start:ez_end] = np.inf

    profile['regimes'] = np.array(regimes, dtype=int)
    profile['cac_ez'] = ez

    return profile



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.pairwise_dist import (
    compute_dist,
    pairwise_dist,
)


def test_pairwise_dist_valid_simple():
    X = [
        np.arange(100),
        np.arange(100),
        np.ones(100),
        np.zeros(100)
    ]
    w = 8
    dists = pairwise_dist(X, w)
    expected = np.array([ 0, 4, 4, 4, 4, 4])
    np.testing.assert_equal(dists, expected)

    # test with MxN np.ndarray
    X = np.array(X)
    dists = pairwise_dist(X, w)
    expected = np.array([ 0, 4, 4, 4, 4, 4])
    np.testing.assert_equal(dists, expected)


def test_pairwise_dist_invalid_params():
    X = [
        np.arange(10),
        np.arange(20)
    ]
    w = 4
    threshold = 0.05
    n_jobs = 1
    with pytest.raises(ValueError) as excinfo:
        pairwise_dist('', w, threshold=threshold, n_jobs=n_jobs)
        assert('X must be array_like!' == str(excinfo.value))

    error = 'threshold must be a float greater than 0 and less'\
                ' than 1'
    with pytest.raises(ValueError) as excinfo:
        pairwise_dist(X, w, threshold=1, n_jobs=n_jobs)
        assert(error == str(excinfo.value))


def test_compute_dist_valid():
    ts = np.arange(100)
    w = 8
    k = 0
    threshold = 0.05
    args = (k, ts, ts, w, threshold)
    result = compute_dist(args)

    assert(result[0] == k)
    assert(result[0] == 0)


# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import logging
import math
import multiprocessing
import sys

import numpy as np

from matrixprofile import cycore

logger = logging.getLogger(__name__)


def mp_pool():
    """
    Utility function to get the appropriate multiprocessing
    handler for Python 2 and 3.
    """
    ctxt = None
    if sys.version_info[0] == 2:
        from contextlib import contextmanager

        @contextmanager
        def multiprocessing_context(*args, **kwargs):
            pool = multiprocessing.Pool(*args, **kwargs)
            yield pool
            pool.terminate()

        ctxt = multiprocessing_context
    else:
        ctxt = multiprocessing.Pool

    return ctxt


def is_array_like(a):
    """
    Helper function to determine if a value is array like.

    Parameters
    ----------
    a : obj
        Object to test.

    Returns
    -------
    True or false respectively.
    """
    return isinstance(a, (list, tuple, np.ndarray))


def is_similarity_join(ts_a, ts_b):
    """
    Helper function to determine if a similarity join is occuring or not.

    Parameters
    ----------
    ts_a : array_like
        Time series A.
    ts_b : array_like, None
        Time series B.

    Returns
    -------
    True or false respectively.
    """
    return is_array_like(ts_a) and is_array_like(ts_b)


def to_np_array(a):
    """
    Helper function to convert tuple or list to np.ndarray.

    Parameters
    ----------
    a : Tuple, list or np.ndarray
        The object to transform.

    Returns
    -------
    The np.ndarray.

    Raises
    ------
    ValueError
        If a is not a valid type.
    """
    if isinstance(a, np.ndarray):
        return a

    if not is_array_like(a):
        raise ValueError('Unable to convert to np.ndarray!')    

    return np.array(a)


def is_one_dimensional(a):
    """
    Helper function to determine if value is one dimensional.

    Parameters
    ----------
    a : array_like
        Object to test.

    Returns
    -------
    True or false respectively.
    """
    return a.ndim == 1


def get_profile_length(ts_a, ts_b, m):
    """
    Determines the profile length based on the provided inputs.

    Parameters
    ----------
    ts_a : array_like
        Time series containing the queries for which to calculate the Matrix Profile.
    ts_b : array_line
        Time series containing the queries for which to calculate the Matrix Profile.
    m : int
        Length of subsequence to compare.
    
    Returns
    -------
    int - the length of the matrix profile.
    """
    return len(ts_a) - m + 1


def find_skip_locations(ts, profile_length, window_size):
    """
    Determines which locations should be skipped based on nan or inf values.

    Parameters
    ----------
    ts : array_like
        Time series containing the queries for which to calculate the Matrix Profile.
    query : array_line
        Time series containing the queries for which to calculate the Matrix Profile.
    window_size : int
        Length of subsequence to compare.
    
    Returns
    -------
    int - the length of the matrix profile.
    """
    skip_loc = np.zeros(profile_length).astype(bool)

    for i in range(profile_length):
        segment = ts[i:i + window_size]
        search = (np.isinf(segment) | np.isnan(segment))
        
        if np.any(search):
            skip_loc[i] = True
        
    return skip_loc
    

def clean_nan_inf(ts):
    """
    Replaces nan and inf values with zeros per matrix profile algorithms.

    Parameters
    ----------
    ts: array_like
        Time series to clean.
    
    Returns
    -------
    np.ndarray - The cleaned time series.

    Raises
    ------
    ValueError
        When the ts is not array like.
    """
    ts = to_np_array(ts)
    search = (np.isinf(ts) | np.isnan(ts))
    ts[search] = 0

    return ts


def is_nan_inf(val):
    """
    Helper function to determine if a given value is NaN or infinite.

    Parameters
    ----------
    val : obj
        The value to test against.

    Returns
    -------
    boolean - True or False respectively.
    """
    return np.isnan(val) or np.isinf(val)


def is_not_nan_inf(val):
    """
    Helper function to determine if a given value is not NaN or infinite.

    Parameters
    ----------
    val : obj
        The value to test against.

    Returns
    -------
    boolean - True or False respectively.
    """
    not_nan = not np.isnan(val)
    not_inf = not np.isinf(val)
    return not_nan and not_inf


def nan_inf_indices(a):
    """
    Helper method to obtain the nan/inf indices of an array.

    Parameters
    ----------
    a : array_like
        The array to test.

    Returns
    -------
    Masked array of indices containing nan/inf.
    """
    return (np.isnan(a)) | (np.isinf(a))


def not_nan_inf_indices(a):
    """
    Helper method to obtain the non-nan/inf indices of an array.

    Parameters
    ----------
    a : array_like
        The array to test.

    Returns
    -------
    Masked array of indices containing not nan/inf.
    """
    return (~nan_inf_indices(a))


def rolling_window(a, window):
    """
    Provides a rolling window on a numpy array given an array and window size.

    Parameters
    ----------
    a : array_like
        The array to create a rolling window on.
    window : int
        The window size.

    Returns
    -------
    Strided array for computation.
    """
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def moving_average(a, window=3):
    """
    Computes the moving average over an array given a window size.

    Parameters
    ----------
    a : array_like
        The array to compute the moving average on.
    window : int
        The window size.

    Returns
    -------
    The moving average over the array.
    """
    return np.mean(rolling_window(a, window), -1)


def moving_std(a, window=3):
    """
    Computes the moving std. over an array given a window size.

    Parameters
    ----------
    a : array_like
        The array to compute the moving std. on.
    window : int
        The window size.

    Returns
    -------
    The moving std. over the array.
    """
    return np.std(rolling_window(a, window), -1)


def moving_avg_std(a, window=3):
    """
    Computes the moving avg and std. over an array given a window size.

    Parameters
    ----------
    a : array_like
        The array to compute the moving std. on.
    window : int
        The window size.

    Returns
    -------
    The moving avg and std. over the array as a tuple.
    (avg, std)
    """
    a = a.astype('d')
    mu, sig = cycore.moving_avg_std(a, window)

    return (np.asarray(mu), np.asarray(sig))


def moving_min(a, window=3):
    """
    Computes the moving minimum over an array given a window size.

    Parameters
    ----------
    a : array_like
        The array to compute the moving min on.
    window : int
        The window size.

    Returns
    -------
    array_like :
        The array of moving minimums.
    """
    return np.min(rolling_window(a, window), axis=1)


def moving_max(a, window=3):
    """
    Computes the moving maximum over an array given a window size.

    Parameters
    ----------
    a : array_like
        The array to compute the moving max on.
    window : int
        The window size.

    Returns
    -------
    array_like :
        The array of moving maximums.
    """
    return np.max(rolling_window(a, window), axis=1)


def moving_median(a, window=3):
    """
    Computes the moving median over an array given a window size.

    Parameters
    ----------
    a : array_like
        The array to compute the moving median on.
    window : int
        The window size.

    Returns
    -------
    array_like :
        The array of moving medians.
    """
    return np.median(rolling_window(a, window), axis=1)


def fft_convolve(ts, query):
    """
    Computes the sliding dot product for query over the time series using
    the quicker FFT convolution approach.

    Parameters
    ----------
    ts : array_like
        The time series.
    query : array_like
        The query.

    Returns
    -------
    array_like - The sliding dot product.
    """
    n = len(ts)
    m = len(query)
    x = np.fft.fft(ts)
    y = np.append(np.flipud(query), np.zeros([1, n - m]))
    y = np.fft.fft(y)
    z = np.fft.ifft(x * y)

    return np.real(z[m - 1:n])


def sliding_dot_product(ts, query):
    """
    Computes the sliding dot product for query over the time series using
    convolution. Note that the result is trimmed due to the computations
    being invalid; the len(query) to len(ts) is kept.

    Parameters
    ----------
    ts : array_like
        The time series.
    query : array_like
        The query.

    Returns
    -------
    array_like - The sliding dot product.
    """
    m = len(query)
    n = len(ts)
    dp = np.convolve(ts, np.flipud(query), mode='full')

    return np.real(dp[m - 1:n])


def distance_profile(prod, ws, data_mu, data_sig, query_mu, query_sig):
    """
    Computes the distance profile for the given statistics.

    Parameters
    ----------
    prod : array_like
        The sliding dot product between the time series and query.
    ws : int
        The window size.
    data_mu : array_like
        The time series moving average.
    data_sig : array_like
        The time series moving standard deviation.
    query_mu : array_like
        The querys moving average.
    query_sig : array_like
        The querys moving standard deviation.


    Returns
    -------
    array_like - The distance profile.
    """
    distance_profile = (
        2 * (ws - (prod - ws * data_mu * query_mu) / (data_sig * query_sig))
    )

    with np.errstate(divide='ignore', invalid='ignore'):
        distance_profile = np.sqrt(np.real(distance_profile))

    return distance_profile


def precheck_series_and_query_1d(ts, query):
    """
    Helper function to ensure we have 1d time series and query.

    Parameters
    ----------
    ts : array_like
        The array to create a rolling window on.
    query : array_like
        The query.

    Returns
    -------
    (np.array, np.array) - The ts and query respectively.

    Raises
    ------
    ValueError
        If ts is not a list or np.array.
        If query is not a list or np.array.
        If ts or query is not one dimensional.
    """
    try:
        ts = to_np_array(ts)
    except ValueError:
        raise ValueError('Invalid ts value given. Must be array_like!')

    try:
        query = to_np_array(query)
    except ValueError:
        raise ValueError('Invalid query value given. Must be array_like!')

    if not is_one_dimensional(ts):
        raise ValueError('ts must be one dimensional!')

    if not is_one_dimensional(query):
        raise ValueError('query must be one dimensional!')

    return (ts, query)


def valid_n_jobs(n_jobs):
    """
    Validates and assigns correct number of cpu cores.

    Parameters
    ----------
    n_jobs : int
        Number of desired cpu cores.
    
    Returns
    -------
    Valid number of cpu cores.
    """
    max_cpus = multiprocessing.cpu_count()
    if n_jobs < 1:
        n_jobs = max_cpus

    if n_jobs > max_cpus:
        n_jobs = max_cpus
    
    return n_jobs


def generate_batch_jobs(profile_length, n_jobs):
    """
    Generates start and end positions for a matrix profile length and number
    of jobs.

    Parameters
    ----------
    profile_length : int
        The length of the matrix profile to compute.
    n_jobs : int
        The number of jobs (cpu cores).
    

    Returns
    -------
    Yielded start and end index for each job.
    """
    batch_size = int(math.ceil(profile_length / n_jobs))

    if batch_size == profile_length:
        yield (0, profile_length)
    else:
        for i in range(n_jobs):
            start = i * batch_size        
            end = (i + 1) * batch_size
            
            if end > profile_length:
                end = profile_length

            yield (start, end)

            if end == profile_length:
                break


def apply_exclusion_zone(exclusion_zone, is_join, window_size, data_length,
    index, distance_profile):
    if exclusion_zone > 0 and not is_join:
        ez_start = np.max([0, index - exclusion_zone])
        ez_end = np.min([data_length - window_size + 1, index + exclusion_zone + 1])
        distance_profile[ez_start:ez_end] = np.inf

    return distance_profile


def pearson_to_euclidean(a, windows):
    """
    Converts an array of Pearson metrics to Euclidean. The array and windows
    should both be row-wise aligned.

    Parameters
    ----------
    a : array_like
        The array of Pearson metrics.
    windows : int or array_like
        The window(s) used to compute the Pearson metrics.

    Returns
    -------
    New array of same dimensions as input array, but with Euclidean distance.
    """
    euc_a = np.full(a.shape, np.inf, dtype='d')
    
    if is_one_dimensional(a):

        window = windows
        if is_array_like(windows):
            window = windows[0]

        is_inf = np.isinf(a)
        euc_a = np.sqrt(2 * window * (1 - a))
    else:
        for window, idx in zip(windows, range(a.shape[0])):
            is_inf = np.isinf(a[idx])
            euc_a[idx] = np.sqrt(2 * window * (1 - a[idx]))
            euc_a[idx][is_inf] = np.inf
    
    return euc_a


def is_pearson_array(a):
    """
    Helper function to determine if an array contains pearson metrics only. It
    is assumed that if the min is >= 0 and max is <= 1, then it is pearson.

    Parameters
    ----------
    a : array_like
        The array to test.

    Returns
    -------
    True or false respectively.
    """
    mask = ~nan_inf_indices(a)
    min_val = a[mask].min()
    max_val = a[mask].max()

    return min_val >= 0 and max_val <= 1


def is_stats_obj(obj):
    """
    Helper function to determine if the current object matches the structure
    that the library enforices for Statistics.

    Parameters
    ----------
    obj : object
        The object to test.
    
    Returns
    -------
    bool :
        True or false respectively.
    """
    return isinstance(obj, dict) and obj.get('class') == 'Statistics'


def is_mp_obj(obj):
    """
    Helper function to determine if the current object matches the structure
    that the library enforices for MatrixProfiles.

    Parameters
    ----------
    obj : object
        The object to test.
    
    Returns
    -------
    bool :
        True or false respectively.
    """
    return isinstance(obj, dict) and obj.get('class') == 'MatrixProfile'


def is_pmp_obj(obj):
    """
    Helper function to determine if the current object matches the structure
    that the library enforices for Pan-MatrixProfiles.

    Parameters
    ----------
    obj : object
        The object to test.
    
    Returns
    -------
    bool :
        True or false respectively.
    """
    return isinstance(obj, dict) and obj.get('class') == 'PMP'


def is_mp_or_pmp_obj(obj):
    """
    Helper function to determine if the current object matches the structure
    that the library enforices for MatrixProfile and Pan-MatrixProfiles.

    Parameters
    ----------
    obj : object
        The object to test.
    
    Returns
    -------
    bool :
        True or false respectively.
    """
    return is_pmp_obj(obj) or is_mp_obj(obj)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import math

import numpy as np

from matrixprofile import core
from matrixprofile.algorithms.cympx import mpx_ab_parallel as cympx_ab_parallel
from matrixprofile.algorithms.mass2 import mass2


def mpdist(ts, ts_b, w, threshold=0.05, n_jobs=1):
    """
    Computes the MPDist between the two series ts and ts_b. For more details
    refer to the paper:

    Matrix Proﬁle XII: MPdist: A Novel Time Series Distance Measure to Allow 
    Data Mining in More Challenging Scenarios. Shaghayegh Gharghabi, 
    Shima Imani, Anthony Bagnall, Amirali Darvishzadeh, Eamonn Keogh. ICDM 2018

    Parameters
    ----------
    ts : array_like
        The time series to compute the matrix profile for.
    ts_b : array_like
        The time series to compare against.
    w : int
        The window size.
    threshold : float, Default 0.05
        The percentile in which the distance is taken from. By default it is
        set to 0.05 based on empircal research results from the paper. 
        Generally, you should not change this unless you know what you are
        doing! This value must be a float greater than 0 and less than 1.
    n_jobs : int, Default = 1
        Number of cpu cores to use.
    
    Returns
    -------
    float : mpdist
        The MPDist.

    """
    ts = core.to_np_array(ts).astype('d')
    ts_b = core.to_np_array(ts_b).astype('d')
    n_jobs = core.valid_n_jobs(n_jobs)

    if not core.is_one_dimensional(ts):
        raise ValueError('ts must be one dimensional!')

    if not core.is_one_dimensional(ts_b):
        raise ValueError('ts_b must be one dimensional!')

    if not isinstance(threshold, float) or threshold <= 0 or threshold >= 1:
        raise ValueError('threshold must be a float greater than 0 and less'\
            ' than 1')

    mp, mpi, mpb, mpib = cympx_ab_parallel(ts, ts_b, w, 0, n_jobs)

    mp_abba = np.append(mp, mpb)
    data_len = len(ts) + len(ts_b)
    abba_sorted = np.sort(mp_abba[~core.nan_inf_indices(mp_abba)])

    distance = np.inf
    if len(abba_sorted) > 0:
        upper_idx = int(np.ceil(threshold * data_len)) - 1
        idx = np.min([len(abba_sorted) - 1, upper_idx])
        distance = abba_sorted[idx]

    return distance


def mass_distance_matrix(ts, query, w):
    """
    Computes a distance matrix using mass that is used in mpdist_vector
    algorithm.

    Parameters
    ----------
    ts : array_like
        The time series to compute the matrix for.
    query : array_like
        The time series to compare against.
    w : int
        The window size.
    
    Returns
    -------
    array_like : dist_matrix
        The MASS distance matrix.

    """
    subseq_num = len(query) - w + 1
    distances = []
    
    for i in range(subseq_num):
        distances.append(np.real(mass2(ts, query[i:i + w])))
    
    return np.array(distances)


def calculate_mpdist(profile, threshold, data_length):
    """
    Computes the MPDist given a profile, threshold and data length. This is
    primarily used for MPDist Vector algorithm.

    Parameters
    ----------
    profile : array_like
        The profile to calculate the mpdist for.
    threshold : float
        The threshold to use in computing the distance.
    data_length : int
        The length of the original data.

    Returns
    -------
    float : mpdist
        The MPDist.

    """
    dist_loc = int(np.ceil(threshold * data_length))
    profile_sorted = np.sort(profile)
    mask = core.not_nan_inf_indices(profile_sorted)
    
    profile_clean = profile_sorted[mask]
    
    if len(profile_clean) < 1:
        distance = np.inf
    elif len(profile_clean) >= dist_loc:
        distance = profile_clean[dist_loc]
    else:
        distance = np.max(profile_clean)
    
    return distance


def mpdist_vector(ts, ts_b, w):
    """
    Computes a vector of MPDist measures.

    Parameters
    ----------
    ts : array_like
        The time series to compute the matrix for.
    ts_b : array_like
        The time series to compare against.
    w : int
        The window size.
    
    Returns
    -------
    array_like : mpdist_vector
        The MPDist vector.

    """
    matrix = mass_distance_matrix(ts, ts_b, w)
    rows, cols = matrix.shape

    # compute row and column minimums
    all_right_hist = matrix.min(axis=0)
    mass_minimums = np.apply_along_axis(core.moving_min, 1, matrix, window=rows)

    # recreate the matrix profile and compute MPDist
    mpdist_length = len(ts) - len(ts_b) + 1
    right_hist_length = len(ts_b) - w + 1
    mpdist_array = np.zeros(mpdist_length)
    left_hist = np.zeros(right_hist_length)
    
    mpdist_array = []
    for i in range(mpdist_length):
        right_hist = all_right_hist[i:right_hist_length + i]
        left_hist = mass_minimums[:, i]
        profile = np.append(left_hist, right_hist)
        mpdist_array.append(calculate_mpdist(profile, 0.05, 2 * len(ts_b)))
    
    return np.array(mpdist_array)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import csv
import gzip
import json
import os

# load urlretrieve for python2 and python3
try:
    from urllib.request import urlretrieve
except:
    from urllib import urlretrieve

import numpy as np
    
DATA_LISTING_URL = 'https://raw.githubusercontent.com/matrix-profile-foundation/mpf-datasets/master/listings.json'
DATA_URL = 'https://raw.githubusercontent.com/matrix-profile-foundation/mpf-datasets/master/{}/{}'
DATA_DIR = os.path.expanduser(os.path.join('~', '.mpf-datasets'))


def create_dirs(path):
    """
    Python 2 and 3 compatible function to make directories. Python 3 has the
    exist_ok option in makedirs, but Python 2 does not.

    Parameters
    ----------
    path : str
        The path to create directories for.

    """
    try:
        os.makedirs(path)
    except:
        pass

    if not os.path.exists(path):
        raise OSError('Unable to create path: {}'.format(path))


def fetch_available(category=None):
    """
    Fetches the available datasets found in
    github.com/matrix-profile-foundation/mpf-datasets github repository.
    Providing a category filters the datasets.
    
    Parameters
    ----------
    category : str, Optional
        The desired category to retrieve datasets by.
    
    Returns
    -------
    list :
        A list of dictionaries containing details about each dataset.
    
    Raises
    ------
    ValueError:
        When a category is provided, but is not found in the listing.

    """
    # download the file and load it
    create_dirs(DATA_DIR)
    output_path = os.path.join(DATA_DIR, 'listings.json')
    result = urlretrieve(DATA_LISTING_URL, output_path)
    
    with open(output_path) as f:
        datasets = json.load(f)
    
    # filter with category
    if category:
        category_found = False
        filtered = []
        
        for dataset in datasets:
            if dataset['category'] == category.lower():
                filtered.append(dataset)
                category_found = True
        
        datasets = filtered
        if not category_found:
            raise ValueError('category {} is not a valid option.'.format(category))
    
    return datasets


def get_csv_indices(fp, is_gzip=False):
    """
    Utility function to provide indices of the datetime dimension and the
    real valued dimensions.
    
    Parameters
    ----------
    fp : str
        The filepath to load.
    is_gzip : boolean, Default False
        Flag to tell if the csv is gzipped.
    
    Returns
    -------
    (dt_index, real_indices) :
        The datetime index and real valued indices.

    """
    first_line = None
    if is_gzip:
        with gzip.open(fp, 'rt') as f:
            first_line = f.readline()
    else:
        with open(fp) as f:
            first_line = f.readline()
    
    dt_index = None
    real_indices = []
    for index, label in enumerate(first_line.split(',')):
        if 'date' in label.lower() or 'time' in label.lower():
            dt_index = index
        else:
            real_indices.append(index)
    
    return dt_index, real_indices


def load(name):
    """
    Loads a MPF dataset by base file name or file name. The match is case 
    insensitive.

    Note
    ----
    An internet connection is required to fetch the data.

    Returns
    -------
    dict :
        The dataset and metadata.

        >>> {
        >>>     'name': The file name loaded,
        >>>     'category': The category the file came from,
        >>>     'description': A description,
        >>>     'data': The real valued data as an np.ndarray,
        >>>     'datetime': The datetime as an np.ndarray
        >>> }

    """
    datasets = fetch_available()
    
    # find the filename in datasets matching either on filename provided or
    # the base name
    filename = None
    category = None
    description = None
    for dataset in datasets:
        base_name = dataset['name'].split('.')[0]
        
        if name.lower() == base_name or name.lower() == dataset['name']:
            filename = dataset['name']
            category = dataset['category']
            description = dataset['description']
    
    if not filename:
        raise ValueError('Could not find dataset {}'.format(name))
        
    # download the file
    output_dir = os.path.join(DATA_DIR, category)
    create_dirs(output_dir)
    output_path = os.path.join(output_dir, filename)
    
    if not os.path.exists(output_path):
        url = DATA_URL.format(category, filename)
        urlretrieve(url, output_path)

    # load the file based on type
    is_txt = filename.endswith('.txt')    
    is_txt_gunzip = filename.endswith('.txt.gz')
    is_csv = filename.endswith('.csv')
    is_csv_gunzip = filename.endswith('.csv.gz')
    
    data = None
    dt_data = None
    if is_txt or is_txt_gunzip:
        data = np.loadtxt(output_path)
    elif is_csv or is_csv_gunzip:
        dt_index, real_indices = get_csv_indices(
            output_path, is_gzip=is_csv_gunzip)

        if isinstance(dt_index, int):
            dt_data = np.genfromtxt(
                output_path,
                dtype='datetime64',
                delimiter=',',
                skip_header=True,
                usecols=[dt_index,]
            )

        data = np.genfromtxt(
            output_path,
            delimiter=',',
            dtype='float64',
            skip_header=True,
            usecols=real_indices
        )
    
    return {
        'name': filename,
        'category': category,
        'description': description,
        'data': data,
        'datetime': dt_data
    }



name = "matrixprofile"

import sys
sys.path.append('../')

from matrixprofile.version import __version__, __version_info__

from matrixprofile.compute import compute
from matrixprofile.visualize import visualize
from matrixprofile.analyze import analyze
from matrixprofile import discover
from matrixprofile import transform
from matrixprofile import utils
from matrixprofile import io
from matrixprofile import algorithms
from matrixprofile import datasets


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os
import tempfile

import pytest

import numpy as np

from matrixprofile import transform
from matrixprofile import compute
from matrixprofile import io

import matrixprofile


def test_av_io_json():
    ts = np.random.uniform(size=1024)
    w = 32

    profile = compute(ts, w)
    profile = transform.apply_av(profile, "default")

    out = os.path.join(tempfile.gettempdir(), 'mp.json')
    io.to_disk(profile, out)

    dprofile = io.from_disk(out)

    keys = set(profile.keys())
    keysb = set(dprofile.keys())

    assert(keys == keysb)

    # check values same
    for k, v in profile.items():
        if isinstance(v, np.ndarray):
            np.testing.assert_equal(v, dprofile[k])
        elif k == 'data':
            pass
        else:
            assert(v == dprofile[k])

    np.testing.assert_equal(profile['data']['ts'], dprofile['data']['ts'])
    np.testing.assert_equal(profile['data']['query'], dprofile['data']['query'])


def test_av_io_MPF():
    ts = np.random.uniform(size=1024)
    w = 32

    profile = compute(ts, w)
    profile = transform.apply_av(profile, "default")

    out = os.path.join(tempfile.gettempdir(), 'mp.mpf')
    io.to_disk(profile, out, format='mpf')

    dprofile = io.from_disk(out)

    keys = set(profile.keys())
    keysb = set(dprofile.keys())

    assert(keys == keysb)

    # check values same
    for k, v in profile.items():
        if isinstance(v, np.ndarray):
            np.testing.assert_equal(v, dprofile[k])
        elif k == 'data':
            pass
        else:
            assert(v == dprofile[k])

    np.testing.assert_equal(profile['data']['ts'], dprofile['data']['ts'])
    np.testing.assert_equal(profile['data']['query'], dprofile['data']['query'])


def test_default_valid():
    ts_arr = [0, 1, 2, 3, 4, 5]
    expect = [1, 1, 1, 1]
    w = 3

    av = transform.make_default_av(ts_arr, w)

    assert(len(av) == len(expect))

    np.testing.assert_almost_equal(av, expect)


def test_complexity_valid():
    ts_arr = [[3., 3., 3., 3., 3., 3.],
              [0., 1., 2., 3., 4., 5.],
              [0., 3., 0., 2., 0., 1.]]
    expect = [[0., 0., 0., 0.],
              [0., 0., 0., 0.],
              [0.47295372330527, 0.32279030890406757, 0.13962038997193682, 0.]]
    w = 3

    for i in range(len(ts_arr)):
        av = transform.make_complexity_av(ts_arr[i], w)

        assert(len(av) == len(expect[i]))

        np.testing.assert_almost_equal(av, expect[i])


def test_meanstd_valid():
    ts_arr = [[3., 3., 3., 3., 3., 3.],
              [-10., 10., -10., 1., -1., 1.],
              [0., 3., 0., 2., 0., 1.]]
    expect = [[0., 0., 0., 0.],
              [0., 0., 1., 1.],
              [0., 0., 1., 1.]]
    w = 3

    for i in range(len(ts_arr)):
        av = transform.make_meanstd_av(ts_arr[i], w)

        assert(len(av) == len(expect[i]))

        np.testing.assert_almost_equal(av, expect[i])


def test_clipping_valid():
    ts_arr = [[3., 3., 3., 3., 3., 3.],
              [0., 1., 2., 3., 4., 5.],
              [0., 3., 0., 2., 0., 1.]]
    expect = [[0., 0., 0., 0.],
              [0., 1., 1., 0.],
              [0., 0.5, 0.5, 1.]]
    w = 3

    for i in range(len(ts_arr)):
        av = transform.make_clipping_av(ts_arr[i], w)

        assert(len(av) == len(expect[i]))

        np.testing.assert_almost_equal(av, expect[i])


def test_default_invalid():
    with pytest.raises(ValueError) as excinfo:
        transform.make_default_av("array", 3)
        assert 'make_default_av expects ts to be array-like' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        transform.make_default_av([[1, 2, 3], [1, 2, 3]], 3)
        assert 'make_default_av expects ts to be one-dimensional' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        transform.make_default_av([1, 2, 3], "window")
        assert 'make_default_av expects window to be an integer' \
            in str(excinfo.value)


def test_complexity_invalid():
    with pytest.raises(ValueError) as excinfo:
        transform.make_complexity_av("array", 3)
        assert 'make_complexity_av expects ts to be array-like' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        transform.make_complexity_av([[1, 2, 3], [1, 2, 3]], 3)
        assert 'make_complexity_av expects ts to be one-dimensional' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        transform.make_complexity_av([1, 2, 3], "window")
        assert 'make_complexity_av expects window to be an integer' \
            in str(excinfo.value)


def test_meanstd_invalid():
    with pytest.raises(ValueError) as excinfo:
        transform.make_meanstd_av("array", 3)
        assert 'make_meanstd_av expects ts to be array-like' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        transform.make_meanstd_av([[1, 2, 3], [1, 2, 3]], 3)
        assert 'make_meanstd_av expects ts to be one-dimensional' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        transform.make_meanstd_av([1, 2, 3], "window")
        assert 'make_meanstd_av expects window to be an integer' \
            in str(excinfo.value)


def test_clipping_invalid():
    with pytest.raises(ValueError) as excinfo:
        transform.make_clipping_av("array", 3)
        assert 'make_clipping_av expects ts to be array-like' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        transform.make_clipping_av([[1, 2, 3], [1, 2, 3]], 3)
        assert 'make_clipping_av expects ts to be one-dimensional' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        transform.make_clipping_av([1, 2, 3], "window")
        assert 'make_clipping_av expects window to be an integer' \
            in str(excinfo.value)


def test_apply_default_av_valid():
    ts = [3., 3., 3., 3., 3., 3., 3., 3.]
    w = 4

    profile = compute(ts, windows=w)
    expect = profile['mp']

    profile = transform.apply_av(profile, "default")

    np.testing.assert_almost_equal(profile['cmp'], expect)


def test_apply_complexity_av_valid():
    ts = [3., 3., 3., 3., 3., 3., 3., 3.]
    w = 4

    profile = compute(ts, windows=w)
    expect = profile['mp'] * 2

    profile = transform.apply_av(profile, "complexity")

    np.testing.assert_almost_equal(profile['cmp'], expect)


def test_apply_meanstd_av_valid():
    ts = [3., 3., 3., 3., 3., 3., 3., 3.]
    w = 4

    profile = compute(ts, windows=w)
    expect = profile['mp'] * 2

    profile = transform.apply_av(profile, "meanstd")

    np.testing.assert_almost_equal(profile['cmp'], expect)


def test_apply_clipping_av_valid():
    ts = [3., 3., 3., 3., 3., 3., 3., 3.]
    w = 4

    profile = compute(ts, windows=w)
    expect = profile['mp'] * 2

    profile = transform.apply_av(profile, "clipping")

    np.testing.assert_almost_equal(profile['cmp'], expect)


def test_apply_custom_av_valid():
    ts = [3., 3., 3., 3., 3., 3., 3., 3.]
    w = 4

    profile = compute(ts, windows=w)
    expect = profile['mp'] * 2

    av = [0., 0., 0., 0., 0.]
    profile = transform.apply_av(profile, "custom", av)

    np.testing.assert_almost_equal(profile['cmp'], expect)


def test_apply_av_invalid():
    ts = [3., 3., 3., 3., 3., 3., 3., 3.]
    w = 4

    with pytest.raises(ValueError) as excinfo:
        transform.apply_av("profile", "default")
        assert 'apply_av expects profile as an MP data structure' \
            in str(excinfo.value)

    profile = compute(ts, windows=w)

    with pytest.raises(ValueError) as excinfo:
        transform.apply_av(profile, "custom", "av")
        assert 'apply_av expects custom_av to be array-like' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        transform.apply_av(profile, "not a parameter")
        assert 'av parameter is invalid' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        transform.apply_av(profile, "custom", [0.9, 0.9, 0.9])
        assert 'Lengths of annotation vector and mp are different' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        transform.apply_av(profile, "custom", [0.5, 0.5, 0.6, 1.2, -0.4])
        assert 'Annotation vector values must be between 0 and 1' \
            in str(excinfo.value)
